from django.db import models
from gkutils.commonutils import FLAGS, PROCESSING_FLAGS, getFlagDefs, ra_to_sex, dec_to_sex, getDateFractionMJD
from math import log10
import sys

# Create your models here.
# This is an auto-generated Django model module.
# You'll have to do the following manually to clean this up:
#     * Rearrange models' order
#     * Make sure each model has one field with primary_key=True
# Feel free to rename the models, but don't rename db_table values or field names.
#
# Also note: You'll have to insert the output of 'django-admin.py sqlcustom [appname]'
# into your database.

# KWS 2010-01-11 Added Postage Stamp Server database tables and added extra columns to
#                tcs_transient_objects and tcs_transient_reobservations.
# KWS 2010-06-11 Added nine new columns to tcs_transient_objects and eight new columns
#                to tcs_transient_reobservations. Added three new columns to
#                tcs_postage_stamp_images. Added new tcs_zoo_requests table.


# 2014-03-11 KWS Added missing group_type definition. We need this to
#                pick out the finder images.
class TcsImageGroups(models.Model):
    """TcsImageGroups.
    """

    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(unique=True, max_length=255)
    group_type = models.IntegerField()
    class Meta:
        """Meta.
        """

        db_table = 'tcs_image_groups'
        managed = False

# 2014-03-12 KWS Added whole MJD property so we don't have to rely on
#                the django template language to truncate the MJD. Note
#                That we can't use the image MJD, because this might be
#                different from that of the Target MJD.
class TcsPostageStampImages(models.Model):
    """TcsPostageStampImages.
    """

    id = models.BigIntegerField(primary_key=True)
    image_type = models.CharField(max_length=20)
    image_filename = models.CharField(max_length=255)
    pss_filename = models.CharField(max_length=255, blank=True)
    mjd_obs = models.FloatField(null=True, blank=True)
    image_group_id = models.ForeignKey(TcsImageGroups, to_field='id', db_column='image_group_id', on_delete=models.CASCADE)
    pss_error_code = models.IntegerField()
    filter = models.CharField(max_length=80, blank=True)
    mask_ratio = models.FloatField(null=True, blank=True)
    mask_ratio_at_core = models.FloatField(null=True, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_postage_stamp_images'
        managed = False

    @property
    def whole_mjd(self):
       """whole_mjd.
       """
       # The second field is always the target MJD
       fields = self.image_filename.split('_')
       m = int(float(fields[1]))
       return m

    @property
    def f(self):
       """f.
       """
       filterNoZeros = None
       if self.filter:
           filterNoZeros = self.filter[0]
       return filterNoZeros

class TcsPostageStampStatusCodes(models.Model):
    """TcsPostageStampStatusCodes.
    """

    id = models.IntegerField(primary_key=True)
    code = models.CharField(max_length=20)
    description = models.CharField(max_length=80, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_postage_stamp_status_codes'
        managed = False

class TcsDetectionLists(models.Model):
    """TcsDetectionLists.
    """

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)
    description = models.CharField(max_length=80, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_detection_lists'
        managed = False

class TcsPostageStampRequests(models.Model):
    """TcsPostageStampRequests.
    """

    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(unique=True, max_length=80)
    pss_id = models.IntegerField(null=True, blank=True)
    download_attempts = models.IntegerField()
    status = models.ForeignKey(TcsPostageStampStatusCodes, to_field='id', db_column='status', on_delete=models.CASCADE)
    created = models.DateTimeField()
    updated = models.DateTimeField(null=True, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_postage_stamp_requests'
        managed = False

class TcsZooRequests(models.Model):
    """TcsZooRequests.
    """

    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(unique=True, max_length=80)
    zoo_id = models.IntegerField(null=True, blank=True)
    download_attempts = models.IntegerField()
    # Re-use the postage stamp status codes.  No need to create new ones.
    status = models.ForeignKey(TcsPostageStampStatusCodes, to_field='id', db_column='status', on_delete=models.CASCADE)
    created = models.DateTimeField()
    updated = models.DateTimeField(null=True, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_zoo_requests'
        managed = False

class TcsCatalogueTables(models.Model):
    """TcsCatalogueTables.
    """

    id = models.IntegerField(primary_key=True)
    table_name = models.CharField(max_length=30)
    description = models.CharField(max_length=60)
    url = models.CharField(max_length=255, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_catalogue_tables'
        managed = False

    def __unicode__(self):
        """__unicode__.
        """
        return self.table_name

class TcsClassificationFlags(models.Model):
    """TcsClassificationFlags.
    """

    flag_id = models.IntegerField(primary_key=True)
    flag_name = models.CharField(max_length=30)
    description = models.CharField(max_length=80)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_classification_flags'
        managed = False

# 2013-05-29 KWS Repurposed the tcs_images table to refer to a triplet set
#                Also added property to return whole MJD, since there is
#                no template tag to facilitate this (amazingly).

class TcsImages(models.Model):
    """TcsImages.
    """

    id = models.BigIntegerField(primary_key=True)
    target = models.CharField(max_length=255)
    ref = models.CharField(max_length=255)
    diff = models.CharField(max_length=255)
    mjd_obs = models.FloatField(null=True, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_images'
        managed = False

    @property
    def whole_mjd(self):
       """whole_mjd.
       """
       m = int(self.mjd_obs)
       return m


# 2011-04-01 KWS Added two new tables for User Defined Lists of objects

class TcsObjectGroupDefinitions(models.Model):
    """TcsObjectGroupDefinitions.
    """

    id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=20)
    description = models.CharField(max_length=80)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_object_group_definitions'
        managed = False


# 2014-01-08 KWS Added tessellation and skycell columns
class TcsCmfMetadata(models.Model):
    """TcsCmfMetadata.
    """

    id = models.BigIntegerField(primary_key=True)
    filename = models.CharField(unique=True, max_length=255, blank=True)
    tessellation = models.CharField(unique=True, max_length=255, blank=True)
    skycell = models.CharField(unique=True, max_length=255, blank=True)
    airmass = models.FloatField(null=True, blank=True)
    angle = models.FloatField(null=True, blank=True)
    cell_bad = models.FloatField(null=True, blank=True)
    cell_biassec = models.CharField(max_length=80, blank=True)
    cell_darktime = models.FloatField(null=True, blank=True)
    cell_gain = models.FloatField(null=True, blank=True)
    cell_readnoise = models.FloatField(null=True, blank=True)
    cell_saturation = models.FloatField(null=True, blank=True)
    cell_trimsec = models.CharField(max_length=80, blank=True)
    cell_varfactor = models.FloatField(null=True, blank=True)
    cell_xbin = models.IntegerField(null=True, blank=True)
    cell_xsize = models.IntegerField(null=True, blank=True)
    cell_xwindow = models.IntegerField(null=True, blank=True)
    cell_ybin = models.IntegerField(null=True, blank=True)
    cell_ysize = models.IntegerField(null=True, blank=True)
    cell_ywindow = models.IntegerField(null=True, blank=True)
    chip_id = models.CharField(max_length=80, blank=True)
    chip_temp = models.CharField(max_length=80, blank=True)
    chip_xsize = models.IntegerField(null=True, blank=True)
    chip_ysize = models.IntegerField(null=True, blank=True)
    dt_phot = models.FloatField(null=True, blank=True)
    exptime = models.FloatField(null=True, blank=True)
    extdata = models.CharField(max_length=80, blank=True)
    exttype = models.CharField(max_length=80, blank=True)
    flimit = models.FloatField(null=True, blank=True)
    fpa_alt = models.FloatField(null=True, blank=True)
    fpa_az = models.FloatField(null=True, blank=True)
    fpa_comment = models.CharField(max_length=80, blank=True)
    fpa_dec = models.CharField(max_length=80, blank=True)
    fpa_detector = models.CharField(max_length=80, blank=True)
    fpa_elevation = models.FloatField(null=True, blank=True)
    fpa_env_dir = models.FloatField(null=True, blank=True)
    fpa_env_humid = models.FloatField(null=True, blank=True)
    fpa_env_temp = models.FloatField(null=True, blank=True)
    fpa_env_wind = models.FloatField(null=True, blank=True)
    fpa_filterid = models.CharField(max_length=80, blank=True)
    fpa_filter = models.CharField(max_length=80, blank=True)
    fpa_focus = models.FloatField(null=True, blank=True)
    fpa_instrument = models.CharField(max_length=80, blank=True)
    fpa_latitude = models.CharField(max_length=80, blank=True)
    fpa_longitude = models.CharField(max_length=80, blank=True)
    fpa_m1tilt = models.FloatField(null=True, blank=True)
    fpa_m1tip = models.FloatField(null=True, blank=True)
    fpa_m1x = models.FloatField(null=True, blank=True)
    fpa_m1y = models.FloatField(null=True, blank=True)
    fpa_m1z = models.FloatField(null=True, blank=True)
    fpa_m2tilt = models.FloatField(null=True, blank=True)
    fpa_m2tip = models.FloatField(null=True, blank=True)
    fpa_m2x = models.FloatField(null=True, blank=True)
    fpa_m2y = models.FloatField(null=True, blank=True)
    fpa_m2z = models.FloatField(null=True, blank=True)
    fpa_object = models.CharField(max_length=80, blank=True)
    fpa_obs_group = models.CharField(max_length=80, blank=True)
    fpa_obs_mode = models.CharField(max_length=80, blank=True)
    fpa_obstype = models.CharField(max_length=80, blank=True)
    fpa_obs = models.CharField(max_length=80, blank=True)
    fpa_pon_time = models.FloatField(null=True, blank=True)
    fpa_posangle = models.FloatField(null=True, blank=True)
    fpa_radecsys = models.CharField(max_length=80, blank=True)
    fpa_ra = models.CharField(max_length=80, blank=True)
    fpa_rotangle = models.FloatField(null=True, blank=True)
    fpa_telescope = models.CharField(max_length=80, blank=True)
    fpa_teltemp_extra = models.FloatField(null=True, blank=True)
    fpa_teltemp_m1cell = models.FloatField(null=True, blank=True)
    fpa_teltemp_m1 = models.FloatField(null=True, blank=True)
    fpa_teltemp_m2 = models.FloatField(null=True, blank=True)
    fpa_teltemp_spider = models.FloatField(null=True, blank=True)
    fpa_teltemp_truss = models.FloatField(null=True, blank=True)
    fpa_temp = models.FloatField(null=True, blank=True)
    fsatur = models.FloatField(null=True, blank=True)
    fwhm_maj = models.FloatField(null=True, blank=True)
    fwhm_min = models.FloatField(null=True, blank=True)
    fw_mj_lq = models.FloatField(null=True, blank=True)
    fw_mj_sg = models.FloatField(null=True, blank=True)
    fw_mj_uq = models.FloatField(null=True, blank=True)
    fw_mn_lq = models.FloatField(null=True, blank=True)
    fw_mn_sg = models.FloatField(null=True, blank=True)
    fw_mn_uq = models.FloatField(null=True, blank=True)
    imageid = models.IntegerField(null=True, blank=True)
    imnaxis1 = models.IntegerField(null=True, blank=True)
    imnaxis2 = models.IntegerField(null=True, blank=True)
    mjd_obs = models.FloatField(unique=True, null=True, blank=True)
    msky_max = models.FloatField(null=True, blank=True)
    msky_min = models.FloatField(null=True, blank=True)
    msky_mn = models.FloatField(null=True, blank=True)
    msky_nx = models.IntegerField(null=True, blank=True)
    msky_ny = models.IntegerField(null=True, blank=True)
    msky_sig = models.FloatField(null=True, blank=True)
    ndet_cr = models.IntegerField(null=True, blank=True)
    ndet_ext = models.IntegerField(null=True, blank=True)
    npsfstar = models.IntegerField(null=True, blank=True)
    nstars = models.IntegerField(null=True, blank=True)
    photcode = models.CharField(max_length=80, blank=True)
    ppsub_input = models.CharField(max_length=80, blank=True)
    ppsub_kernel = models.CharField(max_length=80, blank=True)
    ppsub_reference = models.CharField(max_length=80, blank=True)
    pscamera = models.CharField(max_length=80, blank=True)
    psformat = models.CharField(max_length=80, blank=True)
    sourceid = models.IntegerField(null=True, blank=True)
    zero_pt = models.FloatField(null=True, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_cmf_metadata'
        managed = False

class TcsCrossMatches(models.Model):
    """TcsCrossMatches.
    """

    transient_object_id = models.IntegerField(null=True, blank=True)
    catalogue_object_id = models.CharField(max_length=30, blank=True)
    catalogue_table_id = models.ForeignKey(TcsCatalogueTables, to_field='id', db_column='catalogue_table_id', on_delete=models.CASCADE)
    search_parameters_id = models.IntegerField(null=True, blank=True)
    separation = models.FloatField(null=True, blank=True)
    id = models.IntegerField(primary_key=True)
    z = models.FloatField(null=True, blank=True)
    scale = models.FloatField(null=True, blank=True)
    distance = models.FloatField(null=True, blank=True)
    distance_modulus = models.FloatField(null=True, blank=True)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_cross_matches'
        managed = False

# 2013-06-05 KWS Swapped the order of the columns so that Django-Tables
#                can display the columns in the required order.
# 2013-10-25 KWS Moved observation status so that we can see it in the
#                quickview pages (for eliminating movers)
class TcsTransientObjects(models.Model):
    """TcsTransientObjects.
    """

    id = models.BigIntegerField(primary_key=True)
    followup_id = models.IntegerField(null=True, blank=True)
    followup_flag_date = models.DateField()
    ra_psf = models.FloatField()
    dec_psf = models.FloatField()
    object_classification = models.IntegerField(null=True, blank=True, db_column='object_classification')
    sherlockClassification = models.CharField(max_length=120, db_column='sherlockClassification')
    observation_status = models.CharField(max_length=40, blank=True)
    survey_field = models.CharField(max_length=10, blank=True)
    local_designation = models.CharField(max_length=40, blank=True)
    ps1_designation = models.CharField(max_length=40, blank=True)
    detection_list_id = models.ForeignKey(TcsDetectionLists, null=True, to_field='id', db_column='detection_list_id', on_delete=models.CASCADE)

    ipp_idet = models.IntegerField(null=True, blank=True)
    x_psf = models.FloatField(null=True, blank=True)
    y_psf = models.FloatField(null=True, blank=True)
    x_psf_sig = models.FloatField(null=True, blank=True)
    y_psf_sig = models.FloatField(null=True, blank=True)
    posangle = models.FloatField(null=True, blank=True)
    pltscale = models.FloatField(null=True, blank=True)
    psf_inst_mag = models.FloatField(null=True, blank=True)
    psf_inst_mag_sig = models.FloatField(null=True, blank=True)
    ap_mag = models.FloatField(null=True, blank=True)
    ap_mag_radius = models.FloatField(null=True, blank=True)
    peak_flux_as_mag = models.FloatField(null=True, blank=True)
    cal_psf_mag = models.FloatField(null=True, blank=True)
    cal_psf_mag_sig = models.FloatField(null=True, blank=True)
    sky = models.FloatField(null=True, blank=True)
    sky_sigma = models.FloatField(null=True, blank=True)
    psf_chisq = models.FloatField(null=True, blank=True)
    cr_nsigma = models.FloatField(null=True, blank=True)
    ext_nsigma = models.FloatField(null=True, blank=True)
    psf_major = models.FloatField(null=True, blank=True)
    psf_minor = models.FloatField(null=True, blank=True)
    psf_theta = models.FloatField(null=True, blank=True)
    psf_qf = models.FloatField(null=True, blank=True)
    psf_ndof = models.IntegerField(null=True, blank=True)
    psf_npix = models.IntegerField(null=True, blank=True)
    moments_xx = models.FloatField(null=True, blank=True)
    moments_xy = models.FloatField(null=True, blank=True)
    moments_yy = models.FloatField(null=True, blank=True)
    flags = models.IntegerField(null=True, blank=True)
    n_frames = models.IntegerField(null=True, blank=True)
    padding = models.IntegerField(null=True, blank=True)
    local_comments = models.CharField(max_length=255, blank=True)
    htm20id = models.IntegerField(null=True, db_column='htm20ID', blank=True) # Field name made lowercase.
    htm16id = models.IntegerField(null=True, db_column='htm16ID', blank=True) # Field name made lowercase.
    cx = models.FloatField(null=True, blank=True)
    cy = models.FloatField(null=True, blank=True)
    cz = models.FloatField(null=True, blank=True)
    tcs_cmf_metadata_id = models.ForeignKey(TcsCmfMetadata, to_field='id', db_column='tcs_cmf_metadata_id', on_delete=models.CASCADE)
#   This Foreign key reference to the classification flags table will need to be removed if the object
#   is to be a member of more than one classification.
    tcs_images_id = models.ForeignKey(TcsImages, to_field='id', db_column='tcs_images_id', on_delete=models.CASCADE)
    date_inserted = models.DateTimeField()
    date_modified = models.DateTimeField(null=True, blank=True)
#    object_classification = models.ForeignKey(TcsClassificationFlags, to_field='flag_id', db_column='object_classification', on_delete=models.CASCADE)
    followup_priority = models.IntegerField(null=True, blank=True)
    external_reference_id = models.CharField(max_length=40, blank=True)
    postage_stamp_request_id = models.ForeignKey(TcsPostageStampRequests, null=True, to_field='id', db_column='postage_stamp_request_id', on_delete=models.CASCADE)
    image_group_id = models.ForeignKey(TcsImageGroups, null=True, to_field='id', db_column='image_group_id', on_delete=models.CASCADE)


    # 2010-02-25 KWS Eight new columns added
    followup_counter = models.IntegerField(null=True, blank=True)
    other_designation = models.CharField(max_length=40, blank=True)
    rb_cat = models.FloatField(null=True, blank=True, db_column='classification_confidence')
    rb_pix = models.FloatField(null=True, blank=True, db_column='confidence_factor')
    quality_threshold_pass = models.BooleanField(null=True, blank=True)

    # 2010-06-11 KWS New columns added for diff stats, local magnitude calculations and zoo requests
    locally_calculated_mag = models.FloatField(null=True, blank=True)
    zoo_request_id = models.ForeignKey(TcsZooRequests, null=True, to_field='id', db_column='zoo_request_id', on_delete=models.CASCADE)
    psf_inst_flux = models.FloatField(null=True, blank=True)
    psf_inst_flux_sig = models.FloatField(null=True, blank=True)
    diff_npos = models.IntegerField(null=True, blank=True)
    diff_fratio = models.FloatField(null=True, blank=True)
    diff_nratio_bad = models.FloatField(null=True, blank=True)
    diff_nratio_mask = models.FloatField(null=True, blank=True)
    diff_nratio_all = models.FloatField(null=True, blank=True)

    # 2011-02-24 KWS New column added for Observation Status and Current Trend
    current_trend = models.CharField(max_length=40, blank=True)

    # 2013-12-05 KWS Added processing_status for processing flag defs
    processing_flags = models.IntegerField(null=True, blank=True)

    # 2021-06-09 KWS Added the 'xt' flag (suspected crosstalk)
    xt = models.IntegerField(null=True, blank=True)


    class Meta:
        """Meta.
        """

        db_table = 'tcs_transient_objects'
        managed = False

    @property
    def flags_bin(self):
       """flags_bin.
       """
       a = bin(int(str(self.flags)),32)
       return a

    @property
    def instrument_mag(self):
       """instrument_mag.
       """
       zero_pt = self.tcs_cmf_metadata_id.zero_pt
       exptime = self.tcs_cmf_metadata_id.exptime
       if zero_pt == None or self.psf_inst_mag == None or exptime == None:
           imag = None
       else:
           imag = self.psf_inst_mag + 2.5 * log10(exptime) + zero_pt
       return imag

    @property
    def aperture_mag(self):
       """aperture_mag.
       """
       zero_pt = self.tcs_cmf_metadata_id.zero_pt
       exptime = self.tcs_cmf_metadata_id.exptime
       if zero_pt == None or self.ap_mag == None or exptime == None:
           amag = None
       else:
           amag = self.ap_mag + 2.5 * log10(exptime) + zero_pt
       return amag

    @property
    def calibrated_mag(self):
       """calibrated_mag.
       """
       zero_pt = self.tcs_cmf_metadata_id.zero_pt
       exptime = self.tcs_cmf_metadata_id.exptime
       if zero_pt == None or self.cal_psf_mag == None or exptime == None:
           cmag = None
       else:
           cmag = self.cal_psf_mag + 2.5 * log10(exptime) + zero_pt
       return cmag

    @property
    def ra_sex(self):
        """ra_sex.
        """
        ra_in_sex = ra_to_sex (self.ra_psf)
        return ra_in_sex

    @property
    def dec_sex(self):
        """dec_sex.
        """
        dec_in_sex = dec_to_sex (self.dec_psf)
        return dec_in_sex

    @property
    def ra_in_hours(self):
        """ra_in_hours.
        """
        ra_hours = ra_in_decimal_hours(self.ra_psf)
        return ra_hours

    # 2013-02-05 KWS Add decoded flag bits and break the foreign key relation for object_classification
    @property
    def decode_flag_bits(self):
        """decode_flag_bits.
        """
        object_definition = getFlagDefs(self.object_classification, FLAGS, delimiter = ' ')
        return object_definition

    # 2013-12-05 KWS Add decoded flag bits for processing_status
    @property
    def decode_processing_flags_bits(self):
        """decode_processing_flags_bits.
        """
        pf = getFlagDefs(self.processing_flags, PROCESSING_FLAGS, delimiter = ' ')
        sys.stderr.write('\nFLAG VALUE:%d\n' % self.processing_flags)
        sys.stderr.write('\nFLAG DEFS:%s\n' % pf)
        return pf


class TcsTransientReobservations(models.Model):
    """TcsTransientReobservations.
    """

    ipp_idet = models.IntegerField(null=True, blank=True)
    x_psf = models.FloatField(null=True, blank=True)
    y_psf = models.FloatField(null=True, blank=True)
    x_psf_sig = models.FloatField(null=True, blank=True)
    y_psf_sig = models.FloatField(null=True, blank=True)
    ra_psf = models.FloatField()
    dec_psf = models.FloatField()
    posangle = models.FloatField(null=True, blank=True)
    pltscale = models.FloatField(null=True, blank=True)
    psf_inst_mag = models.FloatField(null=True, blank=True)
    psf_inst_mag_sig = models.FloatField(null=True, blank=True)
    ap_mag = models.FloatField(null=True, blank=True)
    ap_mag_radius = models.FloatField(null=True, blank=True)
    peak_flux_as_mag = models.FloatField(null=True, blank=True)
    cal_psf_mag = models.FloatField(null=True, blank=True)
    cal_psf_mag_sig = models.FloatField(null=True, blank=True)
    sky = models.FloatField(null=True, blank=True)
    sky_sigma = models.FloatField(null=True, blank=True)
    psf_chisq = models.FloatField(null=True, blank=True)
    cr_nsigma = models.FloatField(null=True, blank=True)
    ext_nsigma = models.FloatField(null=True, blank=True)
    psf_major = models.FloatField(null=True, blank=True)
    psf_minor = models.FloatField(null=True, blank=True)
    psf_theta = models.FloatField(null=True, blank=True)
    psf_qf = models.FloatField(null=True, blank=True)
    psf_ndof = models.IntegerField(null=True, blank=True)
    psf_npix = models.IntegerField(null=True, blank=True)
    moments_xx = models.FloatField(null=True, blank=True)
    moments_xy = models.FloatField(null=True, blank=True)
    moments_yy = models.FloatField(null=True, blank=True)
    flags = models.IntegerField(null=True, blank=True)
    n_frames = models.IntegerField(null=True, blank=True)
    padding = models.IntegerField(null=True, blank=True)
    htm20id = models.IntegerField(db_column='htm20ID') # Field name made lowercase.
    htm16id = models.IntegerField(db_column='htm16ID') # Field name made lowercase.
    cx = models.FloatField()
    cy = models.FloatField()
    cz = models.FloatField()
    id = models.BigIntegerField(primary_key=True)
    tcs_cmf_metadata_id = models.ForeignKey(TcsCmfMetadata, to_field='id', db_column='tcs_cmf_metadata_id', on_delete=models.CASCADE)
    tcs_images_id = models.ForeignKey(TcsImages, to_field='id', db_column='tcs_images_id', on_delete=models.CASCADE)
    transient_object_id = models.BigIntegerField()
    date_inserted = models.DateTimeField()
    date_modified = models.DateTimeField(null=True, blank=True)
    local_comments = models.CharField(max_length=255, blank=True)
    postage_stamp_request_id = models.ForeignKey(TcsPostageStampRequests, null=True, to_field='id', db_column='postage_stamp_request_id', on_delete=models.CASCADE)
    image_group_id = models.ForeignKey(TcsImageGroups, null=True, to_field='id', db_column='image_group_id', on_delete=models.CASCADE)

    # 2010-02-25 KWS New column added
    quality_threshold_pass = models.BooleanField(null=True, blank=True)

    # 2010-06-11 KWS New columns added for diff stats and local magnitude calculations
    locally_calculated_mag = models.FloatField(null=True, blank=True)
    psf_inst_flux = models.FloatField(null=True, blank=True)
    psf_inst_flux_sig = models.FloatField(null=True, blank=True)
    diff_npos = models.IntegerField(null=True, blank=True)
    diff_fratio = models.FloatField(null=True, blank=True)
    diff_nratio_bad = models.FloatField(null=True, blank=True)
    diff_nratio_mask = models.FloatField(null=True, blank=True)
    diff_nratio_all = models.FloatField(null=True, blank=True)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_transient_reobservations'
        managed = False

    @property
    def flags_bin(self):
       """flags_bin.
       """
       a = bin(int(str(self.flags)),32)
       return a

    @property
    def instrument_mag(self):
       """instrument_mag.
       """
       zero_pt = self.tcs_cmf_metadata_id.zero_pt
       exptime = self.tcs_cmf_metadata_id.exptime
       if zero_pt == None or self.psf_inst_mag == None or exptime == None:
           imag = None
       else:
           imag = self.psf_inst_mag + 2.5 * log10(exptime) + zero_pt
       return imag

    @property
    def aperture_mag(self):
       """aperture_mag.
       """
       zero_pt = self.tcs_cmf_metadata_id.zero_pt
       exptime = self.tcs_cmf_metadata_id.exptime
       if zero_pt == None or self.ap_mag == None or exptime == None:
           amag = None
       else:
           amag = self.ap_mag + 2.5 * log10(exptime) + zero_pt
       return amag

    @property
    def calibrated_mag(self):
       """calibrated_mag.
       """
       zero_pt = self.tcs_cmf_metadata_id.zero_pt
       exptime = self.tcs_cmf_metadata_id.exptime
       if zero_pt == None or self.cal_psf_mag == None or exptime == None:
           cmag = None
       else:
           cmag = self.cal_psf_mag + 2.5 * log10(exptime) + zero_pt
       return cmag

    @property
    def ra_sex(self):
        """ra_sex.
        """
        ra_in_sex = ra_to_sex (self.ra_psf)
        return ra_in_sex

    @property
    def dec_sex(self):
        """dec_sex.
        """
        dec_in_sex = dec_to_sex (self.dec_psf)
        return dec_in_sex

    @property
    def ra_in_hours(self):
        """ra_in_hours.
        """
        ra_hours = ra_in_decimal_hours(self.ra_psf)
        return ra_hours

# 2011-10-13 KWS Added new model for tcs_ipp_to_cfa_lookup

class TcsIppToCfaLookup(models.Model):
    """TcsIppToCfaLookup.
    """

    transient_object_id = models.BigIntegerField(db_column='transient_object_id', primary_key=True)
    eventID = models.BigIntegerField(null=True, blank=True)
    cfa_designation = models.CharField(max_length=45, db_column='cfa_designation')
    separation = models.FloatField(null=True, blank=True)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_ipp_to_cfa_lookup'
        managed = False


# 2013-10-21 KWS Added new model for tcs_cross_matches_external
class TcsCrossMatchesExternal(models.Model):
    """TcsCrossMatchesExternal.
    """

    id = models.IntegerField(primary_key=True)
    transient_object_id = models.ForeignKey(TcsTransientObjects, to_field='id', db_column='transient_object_id', on_delete=models.CASCADE)
    external_designation = models.CharField(max_length=180)
    type = models.CharField(max_length=120, blank=True)
    host_galaxy = models.CharField(max_length=180, blank=True)
    mag = models.FloatField(null=True, blank=True)
    discoverer = models.CharField(max_length=900, blank=True)
    matched_list = models.CharField(max_length=300)
    other_info = models.CharField(max_length=900, blank=True)
    separation = models.FloatField(null=True, blank=True)
    comments = models.CharField(max_length=900, blank=True)
    url = models.CharField(max_length=900, blank=True)
    host_z = models.FloatField(blank=True, null=True)
    object_z = models.FloatField(blank=True, null=True)
    disc_date = models.DateTimeField(blank=True, null=True)
    disc_filter = models.CharField(max_length=150, blank=True)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_cross_matches_external'
        managed = False

    @property
    def ra_sex(self):
        """ra_sex.
        """
        ra_in_sex = ra_to_sex (self.ra)
        return ra_in_sex

    @property
    def dec_sex(self):
        """dec_sex.
        """
        dec_in_sex = dec_to_sex (self.dec)
        return dec_in_sex


class TcsObjectGroups(models.Model):
    """TcsObjectGroups.
    """

    #id = models.BigIntegerField(primary_key=True)  # This can't be used as an auto increment by Django!
    id = models.AutoField(primary_key=True)
    transient_object_id = models.ForeignKey(TcsTransientObjects, to_field='id', db_column='transient_object_id', on_delete=models.CASCADE)
    object_group_id = models.ForeignKey(TcsObjectGroupDefinitions, to_field='id', db_column='object_group_id', on_delete=models.CASCADE)
    class Meta:
        """Meta.
        """

        db_table = 'tcs_object_groups'
        managed = False

# 2015-11-17 KWS Added tcs_processing_status. We'll use the contents of this table to
#                stop users making updates when the database is locked.
class TcsProcessingStatus(models.Model):
    """TcsProcessingStatus.
    """

    id = models.BigIntegerField(primary_key=True)
    status = models.IntegerField()
    started = models.DateTimeField()
    modified = models.DateTimeField(blank=True, null=True)

    class Meta:
        """Meta.
        """

        managed = False
        db_table = 'tcs_processing_status'

# 2017-05-09 KWS Added TcsObjectComments table. Comments will be inserted here in future.
class TcsObjectComments(models.Model):
    """TcsObjectComments.
    """

    id = models.BigIntegerField(db_column='id', primary_key=True)
    transient_object_id = models.ForeignKey(TcsTransientObjects, to_field='id', db_column='transient_object_id', on_delete=models.CASCADE)
    date_inserted = models.DateTimeField(db_column='date_inserted', blank=False, null=False)
    comment = models.CharField(max_length=768, db_column='comment', blank=True, null=True)
    username = models.CharField(max_length=90, db_column='username', blank=True, null=True)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_object_comments'
        managed = False

# 2017-03-21 KWS Four new database tables for Sherlock and Gravitational Waves.
# 2017-06-20 KWS Dave has updated the definition of sherlock_classifications.
class SherlockClassifications(models.Model):
    """SherlockClassifications.
    """

    transient_object_id = models.OneToOneField(TcsTransientObjects, to_field='id', db_column='transient_object_id', primary_key=True, on_delete=models.CASCADE)
    classification = models.CharField(max_length=45, blank=True, null=True)
    annotation = models.TextField(blank=True, null=True)
    summary = models.CharField(max_length=50, blank=True, null=True)
    matchverified = models.IntegerField(db_column='matchVerified', blank=True, null=True)  # Field name made lowercase.
    developmentcomment = models.CharField(db_column='developmentComment', max_length=100, blank=True, null=True)  # Field name made lowercase.
    datelastmodified = models.DateTimeField(db_column='dateLastModified', blank=True, null=True)  # Field name made lowercase.
    datecreated = models.DateTimeField(db_column='dateCreated', blank=True, null=True)  # Field name made lowercase.
    updated = models.CharField(max_length=45, blank=True, null=True)

    class Meta:
        """Meta.
        """

        managed = False
        db_table = 'sherlock_classifications'


# 2017-06-20 KWS Dave has updated the definition of sherlock_crossmatches.
class SherlockCrossmatches(models.Model):
    """SherlockCrossmatches.
    """

    transient_object_id = models.ForeignKey(TcsTransientObjects, to_field='id', db_column='transient_object_id', on_delete=models.CASCADE)
    catalogue_object_id = models.CharField(max_length=30, blank=True, null=True)
    catalogue_table_id = models.SmallIntegerField(blank=True, null=True)
    separationarcsec = models.FloatField(db_column='separationArcsec', blank=True, null=True)  # Field name made lowercase.
    northseparationarcsec = models.FloatField(db_column='northSeparationArcsec', blank=True, null=True)  # Field name made lowercase.
    eastseparationarcsec = models.FloatField(db_column='eastSeparationArcsec', blank=True, null=True)  # Field name made lowercase.
    id = models.BigIntegerField(primary_key=True)
    z = models.FloatField(blank=True, null=True)
    scale = models.FloatField(blank=True, null=True)
    distance = models.FloatField(blank=True, null=True)
    distance_modulus = models.FloatField(blank=True, null=True)
    photoz = models.FloatField(db_column='photoZ', blank=True, null=True)  # Field name made lowercase.
    photozerr = models.FloatField(db_column='photoZErr', blank=True, null=True)  # Field name made lowercase.
    association_type = models.CharField(max_length=45, blank=True, null=True)
    datecreated = models.DateTimeField(db_column='dateCreated', blank=True, null=True)  # Field name made lowercase.
    physical_separation_kpc = models.FloatField(blank=True, null=True)
    catalogue_object_type = models.CharField(max_length=45, blank=True, null=True)
    catalogue_object_subtype = models.CharField(max_length=45, blank=True, null=True)
    association_rank = models.IntegerField(blank=True, null=True)
    catalogue_table_name = models.CharField(max_length=100, blank=True, null=True)
    catalogue_view_name = models.CharField(max_length=100, blank=True, null=True)
    rank = models.IntegerField(blank=True, null=True)
    rankscore = models.FloatField(db_column='rankScore', blank=True, null=True)  # Field name made lowercase.
    search_name = models.CharField(max_length=100, blank=True, null=True)
    major_axis_arcsec = models.FloatField(blank=True, null=True)
    direct_distance = models.FloatField(blank=True, null=True)
    direct_distance_scale = models.FloatField(blank=True, null=True)
    direct_distance_modulus = models.FloatField(blank=True, null=True)
    radeg = models.FloatField(db_column='raDeg', blank=True, null=True)  # Field name made lowercase.
    decdeg = models.FloatField(db_column='decDeg', blank=True, null=True)  # Field name made lowercase.
    original_search_radius_arcsec = models.FloatField(blank=True, null=True)
    catalogue_view_id = models.IntegerField(blank=True, null=True)
    u = models.FloatField(db_column='U', blank=True, null=True)  # Field name made lowercase.
    uerr = models.FloatField(db_column='UErr', blank=True, null=True)  # Field name made lowercase.
    b = models.FloatField(db_column='B', blank=True, null=True)  # Field name made lowercase.
    berr = models.FloatField(db_column='BErr', blank=True, null=True)  # Field name made lowercase.
    v = models.FloatField(db_column='V', blank=True, null=True)  # Field name made lowercase.
    verr = models.FloatField(db_column='VErr', blank=True, null=True)  # Field name made lowercase.
    r = models.FloatField(db_column='R', blank=True, null=True)  # Field name made lowercase.
    rerr = models.FloatField(db_column='RErr', blank=True, null=True)  # Field name made lowercase.
    i = models.FloatField(db_column='I', blank=True, null=True)  # Field name made lowercase.
    ierr = models.FloatField(db_column='IErr', blank=True, null=True)  # Field name made lowercase.
    j = models.FloatField(db_column='J', blank=True, null=True)  # Field name made lowercase.
    jerr = models.FloatField(db_column='JErr', blank=True, null=True)  # Field name made lowercase.
    h = models.FloatField(db_column='H', blank=True, null=True)  # Field name made lowercase.
    herr = models.FloatField(db_column='HErr', blank=True, null=True)  # Field name made lowercase.
    k = models.FloatField(db_column='K', blank=True, null=True)  # Field name made lowercase.
    kerr = models.FloatField(db_column='KErr', blank=True, null=True)  # Field name made lowercase.
    field_u = models.FloatField(db_column='_u', blank=True, null=True)  # Field renamed because it started with '_'.
    field_uerr = models.FloatField(db_column='_uErr', blank=True, null=True)  # Field name made lowercase. Field renamed because it started with '_'.
    field_g = models.FloatField(db_column='_g', blank=True, null=True)  # Field renamed because it started with '_'.
    field_gerr = models.FloatField(db_column='_gErr', blank=True, null=True)  # Field name made lowercase. Field renamed because it started with '_'.
    field_r = models.FloatField(db_column='_r', blank=True, null=True)  # Field renamed because it started with '_'.
    field_rerr = models.FloatField(db_column='_rErr', blank=True, null=True)  # Field name made lowercase. Field renamed because it started with '_'.
    field_i = models.FloatField(db_column='_i', blank=True, null=True)  # Field renamed because it started with '_'.
    field_ierr = models.FloatField(db_column='_iErr', blank=True, null=True)  # Field name made lowercase. Field renamed because it started with '_'.
    field_z = models.FloatField(db_column='_z', blank=True, null=True)  # Field renamed because it started with '_'.
    field_zerr = models.FloatField(db_column='_zErr', blank=True, null=True)  # Field name made lowercase. Field renamed because it started with '_'.
    field_y = models.FloatField(db_column='_y', blank=True, null=True)  # Field renamed because it started with '_'.
    field_yerr = models.FloatField(db_column='_yErr', blank=True, null=True)  # Field name made lowercase. Field renamed because it started with '_'.
    g = models.FloatField(db_column='G', blank=True, null=True)  # Field name made lowercase.
    gerr = models.FloatField(db_column='GErr', blank=True, null=True)  # Field name made lowercase.
    unkmag = models.FloatField(db_column='unkMag', blank=True, null=True)  # Field name made lowercase.
    unkmagerr = models.FloatField(db_column='unkMagErr', blank=True, null=True)  # Field name made lowercase.
    datelastmodified = models.DateTimeField(db_column='dateLastModified', blank=True, null=True)  # Field name made lowercase.
    updated = models.IntegerField(blank=True, null=True)
    classificationreliability = models.IntegerField(db_column='classificationReliability', blank=True, null=True)  # Field name made lowercase.
    transientabsmag = models.FloatField(db_column='transientAbsMag', blank=True, null=True)  # Field name made lowercase.

    class Meta:
        """Meta.
        """

        managed = False
        db_table = 'sherlock_crossmatches'


class TcsGravityEvents(models.Model):
    """TcsGravityEvents.
    """

    gravity_event_id = models.CharField(primary_key=True, max_length=10)
    gracedb_id = models.CharField(max_length=10)
    mjd = models.FloatField(blank=True, null=True)
    datelastmodified = models.DateTimeField(db_column='dateLastModified', blank=True, null=True)  # Field name made lowercase.
    updated = models.IntegerField(blank=True, null=True)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_gravity_events'
        managed = False


class TcsGravityEventAnnotations(models.Model):
    """TcsGravityEventAnnotations.
    """

    primaryid = models.BigIntegerField(db_column='primaryId', primary_key=True)  # Field name made lowercase.
    transient_object_id = models.ForeignKey(TcsTransientObjects, to_field='id', db_column='transient_object_id', on_delete=models.CASCADE)
    gravity_event_id = models.ForeignKey(TcsGravityEvents, to_field='gravity_event_id', db_column='gravity_event_id', on_delete=models.CASCADE)
    gracedb_id = models.CharField(max_length=10)
    enclosing_contour = models.IntegerField(blank=True, null=True)
    map_name = models.CharField(max_length=30, blank=True, null=True)
    datelastmodified = models.DateTimeField(db_column='dateLastModified', blank=True, null=True)  # Field name made lowercase.
    updated = models.IntegerField(blank=True, null=True)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_gravity_event_annotations'
        managed = False
        unique_together = (('transient_object_id', 'gracedb_id'),)


# 2019-08-09 KWS Added TcsLatestObjectStats
class TcsLatestObjectStats(models.Model):
    """TcsLatestObjectStats.
    """

    id = models.OneToOneField(TcsTransientObjects, on_delete=models.CASCADE, db_column='id', related_name='stats', primary_key=True)
    latest_mjd = models.FloatField(blank=True, null=True)
    latest_mag = models.FloatField(blank=True, null=True)
    latest_filter = models.CharField(max_length=80, blank=True)
    earliest_mjd = models.FloatField(blank=True, null=True)
    earliest_mag = models.FloatField(blank=True, null=True)
    earliest_filter = models.CharField(max_length=80, blank=True)
    ra_avg = models.FloatField(blank=True, null=True)
    dec_avg = models.FloatField(blank=True, null=True)
    catalogue = models.CharField(max_length=60, blank=True)
    catalogue_object_id = models.CharField(max_length=30, blank=True)
    separation = models.FloatField(blank=True, null=True)
    external_crossmatches = models.CharField(max_length=330, blank=True)

    class Meta:
        """Meta.
        """

        managed = False
        db_table = 'tcs_latest_object_stats'

    @property
    def ra_sex(self):
        """ra_sex.
        """
        ra_in_sex = ra_to_sex (self.ra_avg)
        return ra_in_sex

    @property
    def dec_sex(self):
        """dec_sex.
        """
        dec_in_sex = dec_to_sex (self.dec_avg)
        return dec_in_sex

    @property
    def earliest_mjd_date(self):
        """earliest_mjd_date.
        """
        dateFraction = getDateFractionMJD(self.earliest_mjd, delimiter = '-')
        return dateFraction

    @property
    def latest_mjd_date(self):
        """latest_mjd_date.
        """
        dateFraction = getDateFractionMJD(self.latest_mjd, delimiter = '-')
        return dateFraction

# 2019-12-10 KWS Added TcsZooniverseScores table. Comments will be inserted here in future.
class TcsZooniverseScores(models.Model):
    """TcsZooniverseScores.
    """

    id = models.BigIntegerField(db_column='id', primary_key=True)
    transient_object_id = models.ForeignKey(TcsTransientObjects, to_field='id', db_column='transient_object_id', on_delete=models.CASCADE)
    score = models.FloatField(blank=True, null=True)
    user1 = models.CharField(max_length=384, db_column='user1', blank=True, null=True)
    user2 = models.CharField(max_length=384, db_column='user2', blank=True, null=True)
    user3 = models.CharField(max_length=384, db_column='user3', blank=True, null=True)
    date_inserted = models.DateTimeField(db_column='date_inserted', blank=False, null=False)

    class Meta:
        """Meta.
        """

        db_table = 'tcs_zooniverse_scores'
        managed = False



#
# Written by Min-Su Shin
# Department of Astronomy, University of Michigan (2009 - )
# Department of Astrophysical Sciences, Princeton University (2005 - 2009)
#
# You can freely use the code.
#

import numpy
import math

def sky_median_sig_clip(input_arr, sig_fract, percent_fract, max_iter=100, low_cut=True, high_cut=True):
    """Estimating a sky value for a given number of iterations

    @type input_arr: numpy array
    @param input_arr: image data array
    @type sig_fract: float
    @param sig_fract: fraction of sigma clipping
    @type percent_fract: float
    @param percent_fract: convergence fraction
    @type max_iter: integer
    @param max_iter: max. of iterations
    @type low_cut: boolean
    @param low_cut: cut out only low values
    @type high_cut: boolean
    @param high_cut: cut out only high values
    @rtype: tuple
    @return: (sky value, number of iterations)

    """
    work_arr = numpy.ravel(input_arr)
    old_sky = numpy.median(work_arr)
    sig = work_arr.std()
    upper_limit = old_sky + sig_fract * sig
    lower_limit = old_sky - sig_fract * sig
    if low_cut and high_cut:
        indices = numpy.where((work_arr < upper_limit) & (work_arr > lower_limit))
    else:
        if low_cut:
            indices = numpy.where((work_arr > lower_limit))
        else:
            indices = numpy.where((work_arr < upper_limit))
    work_arr = work_arr[indices]
    new_sky = numpy.median(work_arr)
    iteration = 0
    while ((math.fabs(old_sky - new_sky)/new_sky) > percent_fract) and (iteration < max_iter) :
        iteration += 1
        old_sky = new_sky
        sig = work_arr.std()
        upper_limit = old_sky + sig_fract * sig
        lower_limit = old_sky - sig_fract * sig
        if low_cut and high_cut:
            indices = numpy.where((work_arr < upper_limit) & (work_arr > lower_limit))
        else:
            if low_cut:
                indices = numpy.where((work_arr > lower_limit))
            else:
                indices = numpy.where((work_arr < upper_limit))
        work_arr = work_arr[indices]
        new_sky = numpy.median(work_arr)
    return (new_sky, iteration)



def sky_mean_sig_clip(input_arr, sig_fract, percent_fract, max_iter=100, low_cut=True, high_cut=True):
    """Estimating a sky value for a given number of iterations

    @type input_arr: numpy array
    @param input_arr: image data array
    @type sig_fract: float
    @param sig_fract: fraction of sigma clipping
    @type percent_fract: float
    @param percent_fract: convergence fraction
    @type max_iter: integer
    @param max_iter: max. of iterations
    @type low_cut: boolean
    @param low_cut: cut out only low values
    @type high_cut: boolean
    @param high_cut: cut out only high values
    @rtype: tuple
    @return: (sky value, number of iterations)

    """
    work_arr = numpy.ravel(input_arr)
    old_sky = numpy.mean(work_arr)
    sig = work_arr.std()
    upper_limit = old_sky + sig_fract * sig
    lower_limit = old_sky - sig_fract * sig
    if low_cut and high_cut:
        indices = numpy.where((work_arr < upper_limit) & (work_arr > lower_limit))
    else:
        if low_cut:
            indices = numpy.where((work_arr > lower_limit))
        else:
            indices = numpy.where((work_arr < upper_limit))
    work_arr = work_arr[indices]
    new_sky = numpy.mean(work_arr)
    iteration = 0
    while ((math.fabs(old_sky - new_sky)/new_sky) > percent_fract) and (iteration < max_iter) :
        iteration += 1
        old_sky = new_sky
        sig = work_arr.std()
        upper_limit = old_sky + sig_fract * sig
        lower_limit = old_sky - sig_fract * sig
        if low_cut and high_cut:
            indices = numpy.where((work_arr < upper_limit) & (work_arr > lower_limit))
        else:
            if low_cut:
                indices = numpy.where((work_arr > lower_limit))
            else:
                indices = numpy.where((work_arr < upper_limit))
        work_arr = work_arr[indices]
        new_sky = numpy.mean(work_arr)
    return (new_sky, iteration)



def range_from_zscale(input_arr, contrast = 1.0, sig_fract = 3.0, percent_fract = 0.01, max_iter=100, low_cut=True, high_cut=True):
    """Estimating ranges with the zscale algorithm

    @type input_arr: numpy array
    @param input_arr: image data array as sample pixels to derive z-ranges
    @type contrast: float
    @param contrast: zscale contrast which should be larger than 0.
    @type sig_fract: float
    @param sig_fract: fraction of sigma clipping
    @type percent_fract: float
    @param percent_fract: convergence fraction
    @type max_iter: integer
    @param max_iter: max. of iterations
    @type low_cut: boolean
    @param low_cut: cut out only low values
    @type high_cut: boolean
    @param high_cut: cut out only high values
    @rtype: tuple
    @return: (min. value, max. value, number of iterations)

    """
    work_arr = numpy.ravel(input_arr)
    work_arr = numpy.sort(work_arr) # sorting is done.
    max_ind = len(work_arr) - 1
    midpoint_ind = int(len(work_arr)*0.5)
    I_midpoint = work_arr[midpoint_ind]
    print(".. midpoint index ", midpoint_ind, " I_midpoint ", I_midpoint)
    # initial estimation of the slope
    x = numpy.array(list(range(0, len(work_arr)))) - midpoint_ind
    y = numpy.array(work_arr)
    temp = numpy.vstack([x, numpy.ones(len(x))]).T
    slope, intercept = numpy.linalg.lstsq(temp, y)[0]
    old_slope = slope
    print("... slope & intercept ", old_slope, " ", intercept)
    # initial clipping
    sig = y.std()
    upper_limit = I_midpoint + sig_fract * sig
    lower_limit = I_midpoint - sig_fract * sig
    if low_cut and high_cut:
        indices = numpy.where((work_arr < upper_limit) & (work_arr > lower_limit))
    else:
        if low_cut:
            indices = numpy.where((work_arr > lower_limit))
        else:
            indices = numpy.where((work_arr < upper_limit))
    # new estimation of the slope
    x = numpy.array(indices[0]) - midpoint_ind
    y = numpy.array(work_arr[indices])
    temp = numpy.vstack([x, numpy.ones(len(x))]).T
    slope, intercept = numpy.linalg.lstsq(temp, y)[0]
    new_slope = slope
    print("... slope & intercept ", new_slope, " ", intercept)
    iteration = 1
    # to run the iteration, we need more than 50% of the original input array
    while (((math.fabs(old_slope - new_slope)/new_slope) > percent_fract) and (iteration < max_iter)) and (len(y) >= midpoint_ind) :
        iteration += 1
        old_slope = new_slope
        # clipping
        sig = y.std()
        upper_limit = I_midpoint + sig_fract * sig
        lower_limit = I_midpoint - sig_fract * sig
        if low_cut and high_cut:
            indices = numpy.where((work_arr < upper_limit) & (work_arr > lower_limit))
        else:
            if low_cut:
                indices = numpy.where((work_arr > lower_limit))
            else:
                indices = numpy.where((work_arr < upper_limit))
        # new estimation of the slope
        x = numpy.array(indices[0]) - midpoint_ind
        y = work_arr[indices]
        temp = numpy.vstack([x, numpy.ones(len(x))]).T
        slope, intercept = numpy.linalg.lstsq(temp, y)[0]
        new_slope = slope
        print("... slope & intercept ", new_slope, " ", intercept)

    z1 = I_midpoint + (new_slope / contrast) * (0 - midpoint_ind)
    z2 = I_midpoint + (new_slope / contrast) * (max_ind - midpoint_ind)

    return (z1, z2, iteration)



def range_from_percentile(input_arr, low_cut=0.25, high_cut=0.25):
    """Estimating ranges with given percentiles

    @type input_arr: numpy array
    @param input_arr: image data array as sample pixels to derive ranges
    @type low_cut: float
    @param low_cut: cut of low-value pixels
    @type high_cut: float
    @param high_cut: cut of high-value pixels
    @rtype: tuple
    @return: (min. value, max. value)

    """
    work_arr = numpy.ravel(input_arr)
    work_arr = numpy.sort(work_arr) # sorting is done.
    size_arr = len(work_arr)
    low_size = int(size_arr * low_cut)
    high_size = int(size_arr * high_cut)
    
    z1 = work_arr[low_size]
    z2 = work_arr[size_arr - 1 - high_size]

    return (z1, z2)



def histeq(inputArray, num_bins=1024):
    """Performs histogram equalisation of the input numpy array.
    
    @type inputArray: numpy array
    @param inputArray: image data array
    @type num_bins: int
    @param num_bins: number of bins in which to perform the operation (e.g. 1024)
    @rtype: numpy array
    @return: image data array
    
    """        
    
    imageData=numpy.array(inputArray, copy=True)
    
    # histogram equalisation: we want an equal number of pixels in each intensity range
    sortedDataIntensities=numpy.sort(numpy.ravel(imageData))    
    median=numpy.median(sortedDataIntensities)
    
    # Make cumulative histogram of data values, simple min-max used to set bin sizes and range
    dataCumHist=numpy.zeros(num_bins)
    minIntensity=sortedDataIntensities.min()    
    maxIntensity=sortedDataIntensities.max()
    histRange=maxIntensity-minIntensity
    binWidth=histRange/float(num_bins-1)
    for i in range(len(sortedDataIntensities)):
        binNumber=int(math.ceil((sortedDataIntensities[i]-minIntensity)/binWidth))
        addArray=numpy.zeros(num_bins)
        onesArray=numpy.ones(num_bins-binNumber)
        onesRange=list(range(binNumber, num_bins))
        numpy.put(addArray, onesRange, onesArray)
        dataCumHist=dataCumHist+addArray
                
    # Make ideal cumulative histogram
    idealValue=dataCumHist.max()/float(num_bins)
    idealCumHist=numpy.arange(idealValue, dataCumHist.max()+idealValue, idealValue)
    
    # Map the data to the ideal
    for y in range(imageData.shape[0]):
        for x in range(imageData.shape[1]):
        # Get index corresponding to dataIntensity
            intensityBin=int(math.ceil((imageData[y][x]-minIntensity)/binWidth))
            
    # Guard against rounding errors (happens rarely I think)
    if intensityBin<0:
        intensityBin=0
    if intensityBin>len(dataCumHist)-1:
        intensityBin=len(dataCumHist)-1
        
    # Get the cumulative frequency corresponding intensity level in the data
    dataCumFreq=dataCumHist[intensityBin]
            
    # Get the index of the corresponding ideal cumulative frequency
    idealBin=numpy.searchsorted(idealCumHist, dataCumFreq)
    idealIntensity=(idealBin*binWidth)+minIntensity
    imageData[y][x]=idealIntensity

    scale_min = imageData.min()
    scale_max = imageData.max()
    imageData.clip(min=scale_min, max=scale_max)
    imageData = (imageData -scale_min) / (scale_max - scale_min)
    indices = numpy.where(imageData < 0)
    imageData[indices] = 0.0
        
    return imageData



def linear(inputArray, scale_min=None, scale_max=None):
    """Performs linear scaling of the input numpy array.

    @type inputArray: numpy array
    @param inputArray: image data array
    @type scale_min: float
    @param scale_min: minimum data value
    @type scale_max: float
    @param scale_max: maximum data value
    @rtype: numpy array
    @return: image data array
    
    """        
    print("img_scale : linear")
    imageData=numpy.array(inputArray, copy=True)
    
    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()

    imageData.clip(min=scale_min, max=scale_max)
    imageData = (imageData -scale_min) / (scale_max - scale_min)
    indices = numpy.where(imageData < 0)
    imageData[indices] = 0.0
    
    return imageData


def sqrt(inputArray, scale_min=None, scale_max=None):
    """Performs sqrt scaling of the input numpy array.

    @type inputArray: numpy array
    @param inputArray: image data array
    @type scale_min: float
    @param scale_min: minimum data value
    @type scale_max: float
    @param scale_max: maximum data value
    @rtype: numpy array
    @return: image data array
    
    """        
    
    print("img_scale : sqrt")
    imageData=numpy.array(inputArray, copy=True)
    
    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()

    imageData.clip(min=scale_min, max=scale_max)
    imageData = imageData - scale_min
    indices = numpy.where(imageData < 0)
    imageData[indices] = 0.0
    imageData = numpy.sqrt(imageData)
    imageData = imageData / math.sqrt(scale_max - scale_min)
    
    return imageData


def log(inputArray, scale_min=None, scale_max=None):
    """Performs log10 scaling of the input numpy array.

    @type inputArray: numpy array
    @param inputArray: image data array
    @type scale_min: float
    @param scale_min: minimum data value
    @type scale_max: float
    @param scale_max: maximum data value
    @rtype: numpy array
    @return: image data array
    
    """        
    
    print("img_scale : log")
    imageData=numpy.array(inputArray, copy=True)
    
    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()
    factor = math.log10(scale_max - scale_min)
    indices0 = numpy.where(imageData < scale_min)
    indices1 = numpy.where((imageData >= scale_min) & (imageData <= scale_max))
    indices2 = numpy.where(imageData > scale_max)
    imageData[indices0] = 0.0
    imageData[indices2] = 1.0
    try :
        imageData[indices1] = numpy.log10(imageData[indices1])/factor
    except :
        print("Error on math.log10 for ", (imageData[i][j] - scale_min))

    return imageData


def power(inputArray, power_index=3.0, scale_min=None, scale_max=None):
    """Performs power scaling of the input numpy array.

    @type inputArray: numpy array
    @param inputArray: image data array
    @type power_index: float
    @param power_index: power index
    @type scale_min: float
    @param scale_min: minimum data value
    @type scale_max: float
    @param scale_max: maximum data value
    @rtype: numpy array
    @return: image data array
    
    """        
    
    print("img_scale : power")
    imageData=numpy.array(inputArray, copy=True)
    
    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()
    factor = 1.0 / math.pow((scale_max - scale_min), power_index)
    indices0 = numpy.where(imageData < scale_min)
    indices1 = numpy.where((imageData >= scale_min) & (imageData <= scale_max))
    indices2 = numpy.where(imageData > scale_max)
    imageData[indices0] = 0.0
    imageData[indices2] = 1.0
    imageData[indices1] = numpy.power((imageData[indices1] - scale_min), power_index)*factor

    return imageData


def asinh(inputArray, scale_min=None, scale_max=None, non_linear=2.0):
    """Performs asinh scaling of the input numpy array.

    @type inputArray: numpy array
    @param inputArray: image data array
    @type scale_min: float
    @param scale_min: minimum data value
    @type scale_max: float
    @param scale_max: maximum data value
    @type non_linear: float
    @param non_linear: non-linearity factor
    @rtype: numpy array
    @return: image data array
    
    """        
    
    print("img_scale : asinh")
    imageData=numpy.array(inputArray, copy=True)
    
    if scale_min == None:
        scale_min = imageData.min()
    if scale_max == None:
        scale_max = imageData.max()
    factor = numpy.arcsinh((scale_max - scale_min)/non_linear)
    indices0 = numpy.where(imageData < scale_min)
    indices1 = numpy.where((imageData >= scale_min) & (imageData <= scale_max))
    indices2 = numpy.where(imageData > scale_max)
    imageData[indices0] = 0.0
    imageData[indices2] = 1.0
    imageData[indices1] = numpy.arcsinh((imageData[indices1] - scale_min)/non_linear)/factor

    return imageData




from gkutils.commonutils import readGenericDataFile, readATLASddetHeader
import sys, os

doubles = ['mjd', 'ra', 'dec']
strings = ['obs', 'obj','filt']
shorts = ['nx','ny','rad']

# All items are assumed to be float unless otherwise stated in this schema dictionary
schema = {'mjd': 'double',
          'ra': 'double',
          'dec': 'double',
          'obs': 'string',
          'obj': 'string',
          'filt': 'string',
          'nx': 'unsigned short',
          'ny': 'unsigned short',
          'rad': 'int', 
          'psc': 'short', 
          'pcr': 'short', 
          'pno': 'short', 
          'pbn': 'short', 
          'pmv': 'short', 
          'pvr': 'short', 
          'pxt': 'short', 
          'pkn': 'short', 
          'ptr': 'short',
          'det': 'char',
          'dup': 'char'} 

cToMysql = {'float': 'float',
            'int': 'int',
            'short': 'smallint',
            'unsigned short': 'smallint_unsigned',
            'char': 'tinyint',
            'double': 'double',
            'string': 'varchar'}

cToStringConvert = {'float': 'stof',
            'int': 'stoi',
            'short': 'stoi',
            'unsigned short': 'stoi',
            'char': 'stoi',
            'double': 'stod',
            'string': ''}


typeNulls = {'float': {'type':'NULL_FLOAT_DOUBLE', 'operator': '>='},
             'int': {'type':'NULL_INT', 'operator': '=='},
             'short': {'type':'NULL_SHORT', 'operator': '=='},
             'char': {'type':'NULL_CHAR', 'operator': '=='},
             'unsigned short': {'type':'NULL_UNSIGNED_SHORT', 'operator': '=='},
             'double': {'type':'NULL_FLOAT_DOUBLE', 'operator': '>='},
             'string': {'type':'EMPTY_STRING', 'operator': '=='}}

# How do we define each row?

# How do I define which attributes are part of the classification vector?

# The code needs to read a data file and define the following:

# * header database table
# * detections database table

# * header database insert class header
# * header database insert class
# * detections database insert class header
# * detections database insert class

# For the most part we assume all the attributes are floats.
# Exceptions are:
# * ra (double)
# * dec (double)
# * a few integer columns to be defined
# * members of the classification vector, which will be combined into a long int.

def createDBTable(tableName, suffix, keys, engine='MyISAM'):
    """createDBTable.

    Args:
        tableName:
        suffix:
        keys:
        engine:
    """
    print("cat > %s%s.sql <<EOF" % (tableName, suffix))
    print("CREATE TABLE \`%s%s\` (" % (tableName, suffix))
    print("\`id\` bigint unsigned not null auto_increment,     -- autoincrementing detection id")
    if 'metadata' in tableName:
        print("\`filename\` varchar(255),")
    else:
        print("\`atlas_metadata_id\` bigint(20) unsigned NOT NULL, -- a reference to the input file and other information (OBS data)")
        print("\`atlas_object_id\` bigint(20) unsigned NOT NULL,   -- a reference to the unique object which tags this object")
        print("\`det_id\` int unsigned not null,                   -- the internal id of the detection (unique per file)")
    for row in keys:
        try:
            type = schema[row]
        except KeyError as e:
            type = "float"

        print("\`%s\` %s, " % (row, cToMysql[type]))
    if 'metadata' in tableName:
        print("""\`input\` varchar(255),
\`reference\` varchar(255),""")
    else:
        print("""\`date_modified\` datetime,            -- when was the detection modified?
\`image_group_id\` bigint unsigned, -- a reference to the group of images that refers to this detection
\`quality_threshold_pass\` bool,    -- does this detection pass some specified cuts?
\`deprecated\` bool,                -- option to ignore this detection
\`realbogus_factor\` float,         -- machine learning real/bogus value""")
    print("""\`htm16ID\` bigint unsigned,
\`date_inserted\` datetime NOT NULL,   -- when was the detection inserted?
PRIMARY KEY \`key_id\` (\`id\`),""")
    if 'metadata' in tableName:
        print("""UNIQUE KEY \`key_filename_mjd\` (\`filename\`,\`mjd\`),
KEY \`idx_mjd\` (\`mjd\`),
UNIQUE KEY \`key_obs\` (\`obs\`),
KEY \`idx_texp\` (\`texp\`),
KEY \`idx_filename\` (\`filename\`),
KEY \`idx_input\` (\`input\`),
KEY \`idx_reference\` (\`reference\`),""")
    else:
        print("""KEY \`idx_atlas_object_id\` (\`atlas_object_id\`),
KEY \`idx_atlas_metadata_id\` (\`atlas_metadata_id\`),""")
    print("""KEY \`idx_htm16ID\` (\`htm16ID\`),
KEY \`idx_date_inserted\` (\`date_inserted\`),
KEY \`idx_ra_dec\` (\`ra\`,\`dec\`)
) ENGINE=%s;
EOF
""" % engine)


def createCPPClass(classname, suffix, keys):
    """createCPPClass.

    Args:
        classname:
        suffix:
        keys:
    """
    print("cat > %s%s.h <<EOF" % (classname, suffix))
    print("#ifndef %s%s_H_" % (classname.upper(), suffix.upper()))
    print("#define %s%s_H_" % (classname.upper(), suffix.upper()))
    print("struct %s%s" % (classname, suffix))
    print("{")
    if 'metadata' in classname:
        print("    std::string filename;")
        print("    std::string input;")
        print("    std::string reference;")
    else:
        print("    unsigned long long metadata_id;")
        print("    unsigned long long object_id;")
        print("    unsigned int det_id;")

    for row in keys:
        try:
            type = schema[row]
            if type == 'string':
                type = 'std::string'
        except KeyError as e:
            type = "float"

        print("    %s %s;" % (type, row))

    print("    unsigned long long htm16ID;")
    print("};")
    print("#endif /* %s%s_H_ */" % (classname.upper(), suffix.upper()))
    print("EOF")

def createCPPDBClassH(classname, suffix, keys):
    """createCPPDBClassH.

    Args:
        classname:
        suffix:
        keys:
    """
    print("cat > %s%sDB.h <<EOF" % (classname, suffix))
    print("#ifndef %s%sDB_H_" % (classname.upper(), suffix.upper()))
    print("#define %s%sDB_H_" % (classname.upper(), suffix.upper()))

    print("#include <mysql++.h>")
    print("#include \"DataRowDefinitions.h\"")
    print("#include \"%s%s.h\"" % (classname, suffix))
    print("#include \"CommonUtils.h\"")
    print("#include \"NullMagicNumbers.h\"")

    print("class %s%sDB {" % (classname, suffix))
    print("public:")
    print("    %s%sDB();" % (classname, suffix))

    if 'metadata' in classname:
        print("    unsigned long long insertRow(const %s%s& %s, mysqlpp::Query& query, std::string filename);" % (classname, suffix, classname))
    else:
        print("    unsigned long long insertRow(const %s%s& %s, mysqlpp::Query& query);" % (classname, suffix, classname))
    print("    virtual ~%s%sDB();" % (classname, suffix))
    print("private:")

    if 'metadata' in classname:
        print("    mysqlpp::Null<mysqlpp::sql_varchar> row_filename;")
        print("    mysqlpp::Null<mysqlpp::sql_varchar> row_input;")
        print("    mysqlpp::Null<mysqlpp::sql_varchar> row_reference;")
    else:
        print("    mysqlpp::Null<mysqlpp::sql_bigint_unsigned> row_metadata_id;")
        print("    mysqlpp::Null<mysqlpp::sql_bigint_unsigned> row_object_id;")
        print("    mysqlpp::Null<mysqlpp::sql_int_unsigned> row_det_id;")


    for row in keys:
        try:
            type = schema[row]
        except KeyError as e:
            type = "float"

        print("    mysqlpp::Null<mysqlpp::sql_%s> row_%s;" % (cToMysql[type], row))

    print("    mysqlpp::Null<mysqlpp::sql_bigint_unsigned> row_htm16ID;")

    print("};")
    print("#endif /* %s%sDB_H_ */" % (classname.upper(), suffix.upper()))
    print("EOF")

def createCPPDBClassCPP(classname, suffix, keys):
    """createCPPDBClassCPP.

    Args:
        classname:
        suffix:
        keys:
    """

    print("cat > %s%sDB.cpp <<EOF" % (classname, suffix))
    print("/*")
    print("* %s%sDB.cpp" % (classname, suffix))
    print("*")
    print("*  Created on: May 04, 2017")
    print("*      Author: kws")
    print("*")
    print("*/")
    print("#include \"%s%sDB.h\"" % (classname, suffix))

    print("%s%sDB::%s%sDB():" % (classname, suffix, classname, suffix))
    if 'metadata' in classname:
        print("row_filename(mysqlpp::null),")
        print("row_input(mysqlpp::null),")
        print("row_reference(mysqlpp::null),")
    else:
        print("row_metadata_id(mysqlpp::null),")
        print("row_object_id(mysqlpp::null),")
        print("row_det_id(mysqlpp::null),")

    for row in keys:
        print("row_%s(mysqlpp::null)," % row)

    print("""row_htm16ID(mysqlpp::null)
{
    // TODO Auto-generated constructor stub

}""")

    if 'metadata' in classname:
        print("unsigned long long %s%sDB::insertRow(const %s%s& %s, mysqlpp::Query& query, std::string filename)" % (classname, suffix, classname, suffix, classname.replace('s','')))
    else:
        print("unsigned long long %s%sDB::insertRow(const %s%s& %s, mysqlpp::Query& query)" % (classname, suffix, classname, suffix, classname.replace('s','')))

    print("""{

    // Declare a utils object.  Note that the methods in this class should probably be made static.
    // This would negate the need to constantly declare it.""")

    print("    if (NULL_UNSIGNED_LONG == %s.htm16ID) row_htm16ID = mysqlpp::null; else row_htm16ID = %s.htm16ID;" % (classname.replace('s',''), classname.replace('s','')))

    if 'metadata' in classname:
        print("    if (EMPTY_STRING == %s.filename) row_filename = mysqlpp::null; else row_filename = %s.filename;" % (classname.replace('s',''), classname.replace('s','')))
        print("    if (EMPTY_STRING == %s.input) row_input = mysqlpp::null; else row_input = %s.input;" % (classname.replace('s',''), classname.replace('s','')))
        print("    if (EMPTY_STRING == %s.reference) row_reference = mysqlpp::null; else row_reference = %s.reference;" % (classname.replace('s',''), classname.replace('s','')))
    else:
        print("    if (NULL_UNSIGNED_LONG == %s.metadata_id) row_metadata_id = mysqlpp::null; else row_metadata_id = %s.metadata_id;" % (classname.replace('s',''), classname.replace('s','')))
        print("    if (NULL_UNSIGNED_LONG == %s.object_id) row_object_id = mysqlpp::null; else row_object_id = %s.object_id;" % (classname.replace('s',''), classname.replace('s','')))
        print("    if (NULL_INT == %s.det_id) row_det_id = mysqlpp::null; else row_det_id = %s.det_id;" % (classname.replace('s',''), classname.replace('s','')))

    for row in keys:
        try:
            type = schema[row]
        except KeyError as e:
            type = "float"
        print("    if (%s %s %s.%s) row_%s = mysqlpp::null; else row_%s = %s.%s;" % (typeNulls[type]['type'], typeNulls[type]['operator'], classname.replace('s',''), row, row, row, classname.replace('s',''), row))

    print("    query << std::setprecision(20) << \"INSERT INTO atlas_%s%s\" <<" % (classname, suffix))
    print("                         \"(\" <<")
    if 'metadata' in classname:
        print("                         \"filename,\" <<")
        print("                         \"input,\" <<")
        print("                         \"reference,\" <<")
    else:
        print("                         \"atlas_metadata_id,\" <<")
        print("                         \"atlas_object_id,\" <<")
        print("                         \"det_id,\" <<")

    for row in keys:
        print("                         \"\`%s\`,\" <<" % row)

    print("                         \"htm16ID,\" <<")
    print("                         \"date_inserted\" <<")
    print("                         \")\" <<")
    print("                         \" VALUES (\" <<")

    if 'metadata' in classname:
        print("""                         mysqlpp::quote << row_filename << "," <<
                         mysqlpp::quote << row_input << "," <<
                         mysqlpp::quote << row_reference << "," << """)
    else:
        print("""                         row_metadata_id << "," <<
                         row_object_id << "," <<
                         row_det_id << "," << """)
    for row in keys:
        try:
            type = schema[row]
        except KeyError as e:
            type = "float"

        if type == 'string':
            print("                         mysqlpp::quote << row_%s << \",\" <<" % row)
        else:
            print("                         row_%s << \",\" <<" % row)
    print("""                         row_htm16ID << "," <<
                         "now()" <<
                         ");";
    unsigned long long id = 0;
    try
    {
        query.execute();
        // Grab the auto incrementing ID from the database
        id = query.insert_id();
        query.reset();
    }
    catch (const mysqlpp::Exception& er)
    {
        // Catch-all for MySQL++ exceptions
        std::cerr << "Error: " << er.what() << std::endl;
        query.reset();
    }

    return id;

}
""")

    print("%s%sDB::~%s%sDB() {" % (classname, suffix, classname, suffix))
    print("""
    // TODO Auto-generated destructor stub
}
""")
    print("EOF")

def createIngesterH(suffix, headerKeys, dataKeys):
    """createIngesterH.

    Args:
        suffix:
        headerKeys:
        dataKeys:
    """

    print("cat > DetectionCrossmatcher%s.h <<EOF" % (suffix))
    print("/*")
    print(" * DetectionCrossmatcher%s.h" % (suffix))
    print(" *")
    print(" *  Created on: May 4, 2017")
    print(" *      Author: kws")
    print(" */")
    print("")
    print("#ifndef DETECTIONCROSSMATCHER%s_H_" % (suffix.upper()))
    print("#define DETECTIONCROSSMATCHER%s_H_" % (suffix.upper()))
    print("")
    print("#include <cmath>")
    print("#include <mysql++.h>")
    print("//#include <fstream>")
    print("//#include <iostream>")
    print("")
    print("#include \"CommonUtils.h\"")
    print("#include \"ReadDDTFile.h\"")
    print("")
    print("#include \"Config.h\"")
    print("#include \"DataRowDefinitions.h\"")
    print("#include \"detections%s.h\"" % (suffix))
    print("#include \"metadata%s.h\"" % (suffix))
    print("#include \"DataRowDefinitions.h\"")
    print("#include \"NullMagicNumbers.h\"")
    print("#include \"ReadDDTFile.h\"")
    print("#include \"detections%sDB.h\"" % (suffix))
    print("#include \"metadata%sDB.h\"" % (suffix))
    print("#include \"TransientObjectTableDB.h\"")
    print("#include \"TransientMomentsDB.h\"")
    print("")
    print("// 2014-12-01 KWS Need to inspect the config values.")
    print("")
    print("")
    print("// Transient Classification Flag Designations - over-engineered for 32 exclusive categories")
    print("// and any combination therein.  This has been created to avoid the need to change table")
    print("// structures when a new category is introduced.")
    print("")
    print("")
    print("class DetectionCrossmatcher%s {" % (suffix))
    print("public:")
    print("    DetectionCrossmatcher%s();" % (suffix))
    print("    //int classifyTransients(string dbUser, string dbPass, string dbName, string dbHost, string filename);")
    print("    int crossmatcher(std::map<std::string, DatabaseMember> dbmap, struct FlagsAndRecurrenceRadius frr, std::string filename);")
    print("    virtual ~DetectionCrossmatcher%s();" % (suffix))
    print("")
    print("private:")
    print("    int writeDetectionRow(std::ofstream& outputFile, const metadata%s& metadata, const detections%s& detection);" % (suffix, suffix))
    print("    int createNewObject(ObjectRow& object, detections%s& detection, metadata%s& metadata, MomentsRow& moments, TransientObjectTableDB& objectDB, detections%sDB& detectionDB, TransientMomentsDB& momentsDB, CommonUtils& common, bool writeFile, bool ingestMoments, ofstream& outputFile, mysqlpp::Query& query);" % (suffix, suffix, suffix))
    print("    unsigned int transient_classification_flags;")
    print("    int checkStarProximity(detections%s& detection, mysqlpp::Query& query, mysqlpp::StoreQueryResult& resultSet, CommonUtils& common, struct FlagsAndRecurrenceRadius frr);" % (suffix))
    print("};")
    print("")
    print("#endif /* DETECTIONCROSSMATCHER%s_H_ */" % (suffix.upper()))
    print("EOF")

def createIngesterCPP(suffix, headerKeys, dataKeys):
    """createIngesterCPP.

    Args:
        suffix:
        headerKeys:
        dataKeys:
    """

    print("cat > DetectionCrossmatcher%s.cpp <<EOF" % (suffix))
    print("/*")
    print(" * DetectionCrossmatcher%s.cpp" % (suffix))
    print(" *")
    print(" *  Created on: May 04, 2017")
    print(" *      Author: kws")
    print(" *")
    print(" */")
    print("")
    print("#include \"DetectionCrossmatcher%s.h\"" % (suffix))
    print("")
    print("")
    print("using namespace std;")
    print("using namespace mysqlpp;")
    print("")
    print("DetectionCrossmatcher%s::DetectionCrossmatcher%s():" % (suffix, suffix))
    print("transient_classification_flags(0)")
    print("{")
    print("")
    print("    // TODO Auto-generated constructor stub")
    print("")
    print("}")
    print("")
    print("DetectionCrossmatcher%s::~DetectionCrossmatcher%s() {" % (suffix, suffix))
    print("    // TODO Auto-generated destructor stub")
    print("}")
    print("")
    print("int DetectionCrossmatcher%s::crossmatcher(map<string, DatabaseMember> dbmap, struct FlagsAndRecurrenceRadius frr, string filename)" % (suffix))
    print("{")
    print("    CommonUtils common;")
    print("    string strippedFilename = common.extractFilename(filename);")
    print("")
    print("    string dbUser = dbmap[\"local\"].username;")
    print("    string dbPass = dbmap[\"local\"].password;")
    print("    string dbName = dbmap[\"local\"].database;")
    print("    string dbHost = dbmap[\"local\"].hostname;")
    print("")
    print("    bool skipHTM = frr.skipHTM;")
    print("    bool writeFile = frr.writeFile;")
    print("    double localTransientSearchRadius = frr.recurrenceRadius;")
    print("    double vStarRadius = frr.vStarRadius;")
    print("    unsigned int maxDetections = frr.maxDetections;")
    print("    bool diffDetections = frr.diffDetections;")
    print("    bool ingestMoments = frr.ingestMoments;")
    print("")
    print("    cout << \"Recurrence Radius = \" << localTransientSearchRadius << endl;")
    print("    cout << \"Variable Star Radius = \" << vStarRadius << endl;")
    print("    cout << \"Max Number of Detections = \" << maxDetections << endl;")
    print("")
    print("    // -----------------------------------------------------------------")
    print("")
    print("    // Temporarily use a YAML setting to turn on or off the HTM code.")
    print("    // With HTM turned off, the database is doing a simple insert.")
    print("    if (skipHTM)")
    print("    {")
    print("        cerr << \"Skip HTM cone search - straight insert only\" << endl;")
    print("    }")
    print("    else")
    print("    {")
    print("        cerr << \"Enabled HTM cone search\" << endl;")
    print("    }")
    print("")
    print("")
    print("")
    print("    // Load the DDC Header file into a map and the data into a vector of maps")
    print("    bool ddet = true;")
    print("    string ddetfilenameheader;")



    print("    ReadDDTFile *%s;" % (suffix))
    print("    map<string,string>* %sHeader;" % (suffix))
    print("    vector<map<string,string>*> %sData;" % (suffix))
    print("    %s = new ReadDDTFile(filename);" % (suffix))
    print("")
    print("    // Classification vector")
    print("    // Pvr = \"variable\" means stationary and (very) long persistance: star, AGN, etc.")
    print("    // Ptr = \"transient\" means short persistence, possibly moving")
    print("    // Pmv = \"moving\" uses trailing or other information")
    print("    // Pkn = \"known asteroid\" from catalogs")
    print("    // Psc = \"scar\" means clutter from imperfect star subtration or bleeds")
    print("    // Pxt = \"xtalk\" : x,y sensitive to stars at (x+n*1320),y and (x+n*1320),(10560-y)")
    print("    // Pbn = \"burn\" is a persistence trail away from serial register")
    print("    // Pno = \"noise\" is generic noise")
    print("    // Pcr = \"cosmic\" is cosmic ray")
    print("")
    print("    // Should read the below from a config file!")
    print("    std::set<std::string> cvector;")
    print("    cvector.insert(\"pvr\");")
    print("    cvector.insert(\"ptr\");")
    print("    cvector.insert(\"psc\");")
    print("    cvector.insert(\"pxt\");")
    print("    cvector.insert(\"pbn\");")
    print("    cvector.insert(\"pno\");")
    print("    cvector.insert(\"pcr\");")
    print("")
    print("    // Add the classification vector contents. They should add up to <= 999")
    print("    int classification_vector = 0;")
    print("    unsigned long long htm16ID;")
    print("")
    print("    if (diffDetections)")
    print("    {")
    print("        %sHeader = %s->getHeader();" % (suffix, suffix))
    print("        if (%sHeader)" % (suffix))
    print("        {")
    print("            for (map<string,string>::iterator i = %sHeader->begin(); i != %sHeader->end(); ++i)" % (suffix, suffix))
    print("            {")
    print("                cout << i->first << \": \" << i->second << endl; ")
    print("            }")
    print("        }")
    print("    }")
    print("    %sData = %s->getData();" % (suffix, suffix))
    print("")
    print("    long nrows = %sData.size();" % (suffix))
    print("")
    print("    cout << \"Number of rows to ingest = \" << nrows << endl;")
    print("")
    print("    if (nrows == 0)")
    print("    {")
    print("        cerr << \"No rows to ingest!\" << endl;")
    print("        return -1;")
    print("    }")
    print("")
    print("    if (maxDetections && nrows > maxDetections)")
    print("    {")
    print("        cerr << \"Too many rows (\" << nrows << \") to ingest from \" << filename << \"!  Possible bad data.\" << endl;")
    print("        return -1;")
    print("    }")
    print("")
    print("")
    print("    // Set the \"header\" row.  The TSV contains this info on every row.")
    print("")
    print("    metadata%s metadata;" % (suffix))
    print("")
    print("    metadata.filename = filename;")
    print("")
    print("    metadata.input = \"\";")
    print("    metadata.reference = \"\";")
    print("")
    print("    if (diffDetections)")
    print("    {")
    print("        try")
    print("        {")
    for row in headerKeys:
        try:
            type = schema[row]
        except KeyError as e:
            type = "float"

        cToString = cToStringConvert[type]
        if cToString:
            openType = cToString + '('
            closeType = ')'
        else:
            openType = ''
            closeType = ''

        print("            metadata.%s = %s%sHeader->at(\"%s\")%s;" % (row, openType, suffix, row, closeType))
    print("            metadata.htm16ID = common.generate_htm_id(16,metadata.ra,metadata.dec);")
    print("        }")
    print("        catch (const std::out_of_range& oor)")
    print("        {")
    print("            cerr << \"Mandatory key not found error: \" << oor.what() << endl;")
    print("            cerr << \"Cannot continue...\" << endl;")
    print("            return -1;")
    print("        }")
    print("    }")
    print("")
    print("")
    print("    ofstream outputFile;")
    print("")
    print("    if (writeFile)")
    print("    {")
    print("        cerr << \"Write a detections file.\" << endl;")
    print("        outputFile.open((filename + \".out\").c_str());")
    print("        // Write the header")
    print("        outputFile <<")
    print("        \"metadata_id\" << \"\\t\" <<")
    print("        \"object_id\" << \"\\t\" <<")
    print("        \"det_id\" << \"\\t\" <<")

    for row in dataKeys:
        print("        \"%s\" << \"\\t\" <<" % row)

    print("        \"htm16ID\" <<")
    print("        endl;")
    print("    }")
    print("    else")
    print("    {")
    print("        cerr << \"Inserting detections into the database.\" << endl;")
    print("    }")
    print("")
    print("")
    print("    string catalogue_to_search = \"atlas_diff_objects\";")
    print("")
    print("    // Star catalogs")
    print("    string stars_ps1 = \"tcs_cat_ps1_ubercal_stars\";")
    print("    string stars_sdss_dr9 = \"tcs_cat_sdss_dr9_stars\";")
    print("    string stars_guide_star = \"tcs_guide_star_cat\";")
    print("    string stars_2mass = \"tcs_2mass_psc_cat\";")
    print("")
    print("    unsigned long long metadata_id; // The ID of the FITS header")
    print("")
    print("    try {")
    print("        mysqlpp::Connection conn(true);  // 2015-02-06 KWS I misunderstood the value of false and true: false = Don't throw exceptions, true = throw them.")
    print("")
    print("        if (!conn.connect(dbName.c_str(), dbHost.c_str(), dbUser.c_str(), dbPass.c_str()))")
    print("        {")
    print("            cerr << \"Failed to connect to database.  Exiting.\" << endl;")
    print("            cerr << \"NOTE: Enter \\\"\\\" to denote a blank db password.\" << endl;")
    print("            return -1;")
    print("        }")
    print("")
    print("        // Pass the Query manipulator around by reference.")
    print("        mysqlpp::Query query = conn.query();")
    print("")
    print("        // First we're going to insert the file metadata")
    print("        metadata%sDB header_row;" % (suffix))
    print("        TransientObjectTableDB objectDB;")
    print("        detections%sDB detectionDB;" % (suffix))
    print("        TransientMomentsDB momentsDB;")
    print("")
    print("        metadata_id = header_row.insertRow(metadata, query, strippedFilename);")
    print("")
    print("        if (metadata_id == 0)")
    print("        {")
    print("            // Something went wrong with the insert of the header.  Abort processing of this file!")
    print("            cerr << \"Failed to insert header.  Skipping file processing for \" << strippedFilename << \".\" << endl;")
    print("            return -1;")
    print("        }")
    print("")
    print("        // Column list for cone searching (commented out columns below are examples)")
    print("        std::list<string> columns;")
    print("        columns.push_back(\"*\");")
    print("")
    print("")
    print("        // Now iterate through the detections")
    print("")
    print("        detections%s detection;" % (suffix))
    print("        ObjectRow object;")
    print("        MomentsRow moments;")
    print("        mysqlpp::StoreQueryResult resultSet;")
    print("")
    print("        unsigned int counter = 0;")
    print("        for (vector< map<string,string>* >::iterator p = %sData.begin(); p != %sData.end(); ++p)" % (suffix, suffix))
    print("        {")
    print("            classification_vector = 0;")
    print("            // Is our key one of the classification vectors - if so, lets add them up")
    print("            for (map<string,string>::iterator i = (*p)->begin(); i != (*p)->end(); ++i)")
    print("            {")
    print("                if (cvector.find(i->first) != cvector.end())")
    print("                {")
    print("                    cout << i->first << \": \" << i->second << \" \";")
    print("                    classification_vector += stoi(i->second);")
    print("                }")
    print("            }")
    print("            cout << endl;")
    print("")
    print("            try")
    print("            {")
    for row in dataKeys:
        try:
            type = schema[row]
        except KeyError as e:
            type = "float"

        cToString = cToStringConvert[type]
        if cToString:
            openType = cToString + '('
            closeType = ')'
        else:
            openType = ''
            closeType = ''
        print("                detection.%s = %s(*p)->at(\"%s\")%s;" % (row, openType, row, closeType))
    print("            }")
    print("            catch (const std::out_of_range& oor)")
    print("            {")
    print("                cerr << \"Mandatory key not found error: \" << oor.what() << endl;")
    print("                cerr << \"Cannot continue...\" << endl;")
    print("                return -1;")
    print("            }")
    print("")
    print("            detection.htm16ID = common.generate_htm_id(16,detection.ra,detection.dec);")
    print("")
    print("            //int")
    print("            if (diffDetections)")
    print("                detection.det_id = counter;")
    print("            else")
    print("                detection.det_id = stoi((*p)->at(\"id\"));")
    print("")
    print("            // 2016-03-18 KWS Ingest the moments data.  These will be stored in a")
    print("            //                separate database table which will link to the parent")
    print("            //                detection")
    print("            if (ingestMoments)")
    print("            {")
    print("                try")
    print("                {")
    print("                    moments.x = stof((*p)->at(\"x\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.x = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.y = stof((*p)->at(\"y\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.y = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.cx = stof((*p)->at(\"cx\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.cx = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.cy = stof((*p)->at(\"cy\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.cy = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.xpk = stof((*p)->at(\"xpk\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.xpk = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.ypk = stof((*p)->at(\"ypk\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.ypk = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.rKron = stof((*p)->at(\"rKron\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.rKron = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.rVar = stof((*p)->at(\"rVar\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.rVar = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.flux = stof((*p)->at(\"flux\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.flux = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Crad = stof((*p)->at(\"Crad\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Crad = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Qrad = stof((*p)->at(\"Qrad\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Qrad = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Trad = stof((*p)->at(\"Trad\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Trad = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q2 = stof((*p)->at(\"Q2\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q2 = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q2cos = stof((*p)->at(\"Q2cos\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q2cos = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q2sin = stof((*p)->at(\"Q2sin\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q2sin = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q3cos = stof((*p)->at(\"Q3cos\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q3cos = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q3sin = stof((*p)->at(\"Q3sin\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q3sin = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q1cos = stof((*p)->at(\"Q1cos\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q1cos = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("                try")
    print("                {")
    print("                    moments.Q1sin = stof((*p)->at(\"Q1sin\"));")
    print("                }")
    print("                catch (const std::out_of_range& oor)")
    print("                {")
    print("                    cerr << \"Optional key not found error: \" << oor.what() << endl;")
    print("                    moments.Q1sin = NULL_FLOAT_DOUBLE;")
    print("                }")
    print("            }")
    print("")
    print("")
    print("            // Insert the detection, but before doing so, we need to cone search.")
    print("")
    print("            detection.metadata_id = metadata_id;")
    print("            unsigned long long object_id = 0;")
    print("            unsigned long long detection_id = 0;")
    print("            unsigned long long moments_id = 0;")
    print("")
    print("            // Pick up John Tonry's Indexing ID. We'll use M = 200 (18 arcsec area grid size).")
    print("            // This value of M should also be < 2^31 (signed int).")
    print("            int M = 200;")
    print("")
    print("            // Do HTM cone search but using HTM only.  The results are NOT")
    print("            // refined, so we need to refine them ourselves.")
    print("            if (skipHTM)")
    print("            {")
    print("                createNewObject(object, detection, metadata, moments, objectDB, detectionDB, momentsDB, common, writeFile, ingestMoments, outputFile, query);")
    print("            }")
    print("            else")
    print("            {")
    print("                common.htmCircle(16, detection.ra, detection.dec, localTransientSearchRadius, catalogue_to_search, columns, query);")
    print("")
    print("                resultSet = query.store();")
    print("                query.reset();")
    print("")
    print("                if (resultSet.num_rows() == 0)")
    print("                {")
    print("                    // Find out if it's near a star")
    print("                    object.classification = NULL_INT;")
    print("                    if (frr.matchStars)")
    print("                        object.classification = checkStarProximity(detection, query, resultSet, common, frr);")
    print("                    createNewObject(object, detection, metadata, moments, objectDB, detectionDB, momentsDB, common, writeFile, ingestMoments, outputFile, query);")
    print("                    // Nope - no objects - therefore we need to create a new object.")
    print("                }")
    print("                else")
    print("                {")
    print("                    // We have some results but are they within the required radius?")
    print("                    bool foundObject = false;")
    print("                    for (size_t j = 0; j < resultSet.num_rows(); j++)")
    print("                    {")
    print("                        double separation = 0;")
    print("                        separation = common.calculate_separation(detection.ra, detection.dec, resultSet[j][\"ra\"],resultSet[j][\"dec\"]);")
    print("")
    print("                        if (separation < localTransientSearchRadius)")
    print("                        {")
    print("                            foundObject = true;")
    print("                            // Currently, in Pan-STARRS, we add a new detection row for")
    print("                            // every object within the search radius.  Do we want to do")
    print("                            // that here??")
    print("                            object_id = resultSet[j][\"id\"];")
    print("                            detection.object_id = object_id;")
    print("")
    print("                            if (writeFile)")
    print("                            {")
    print("                                int status = writeDetectionRow(outputFile, metadata, detection);")
    print("                            }")
    print("                            else")
    print("                            {")
    print("                                // Need to insert moment info here with detection_id")
    print("                                detection_id = detectionDB.insertRow(detection, query);")
    print("                                if (ingestMoments && detection_id > 0)")
    print("                                {")
    print("                                    moments.detection_id = detection_id;")
    print("                                    moments_id = momentsDB.insertRow(moments, query);")
    print("                                }")
    print("                            }")
    print("")
    print("                            //detection_id = detectionDB.insertRow(detection, query);")
    print("                            if (detection_id == 0 && !writeFile)")
    print("                            {")
    print("                                // Something went wrong - we want the detection insert to always succeed.")
    print("                                cerr << \"Failed to insert object.\" << endl;")
    print("                                //return -1;")
    print("                            }")
    print("                            if (ingestMoments)")
    print("                            {")
    print("                                if (moments_id == 0 && !writeFile)")
    print("                                {")
    print("                                    // Something went wrong - we want the detection insert to always succeed.")
    print("                                    cerr << \"Failed to insert moments information.\" << endl;")
    print("                                    //return -1;")
    print("                                }")
    print("                            }")
    print("                        }")
    print("                    }")
    print("                    if (!foundObject)")
    print("                    {")
    print("                        object.classification = NULL_INT;")
    print("                        if (frr.matchStars)")
    print("                            object.classification = checkStarProximity(detection, query, resultSet, common, frr);")
    print("                        createNewObject(object, detection, metadata, moments, objectDB, detectionDB, momentsDB, common, writeFile, ingestMoments, outputFile, query);")
    print("                        // Nope - no objects within search radius - therefore we need to create a new object.")
    print("                    }")
    print("")
    print("                }")
    print("")
    print("            }")
    print("")
    print("")
    print("            counter++;")
    print("        }")
    print("")
    print("")
    print("    } catch (mysqlpp::BadQuery er) { // handle any connection or")
    print("        // query errors that may come up")
    print("        std::cerr << \"Error: \" << er.what() << std::endl;")
    print("        return -1;")
    print("    } catch (const mysqlpp::BadConversion& er) {")
    print("        // Handle bad conversions")
    print("        std::cerr << \"Conversion error: \" << er.what() << std::endl")
    print("                << \"\\tretrieved data size: \" << er.retrieved")
    print("                << \", actual size: \" << er.actual_size << std::endl;")
    print("        return -1;")
    print("    } catch (const mysqlpp::Exception& er) {")
    print("        // Catch-all for any other MySQL++ exceptions")
    print("        std::cerr << \"Catch-all Error: \" << er.what() << std::endl;")
    print("        return -1;")
    print("    }")
    print("")
    print("    if (writeFile)")
    print("    {")
    print("        outputFile.close();")
    print("    }")
    print("")
    print("    // Happy with results.")
    print("    return 0;")
    print("")
    print("}")
    print("")
    print("// Write out the detection row into a previously opened file.")
    print("int DetectionCrossmatcher%s::writeDetectionRow(ofstream& outputFile, const metadata%s& metadata, const detections%s& detection)" % (suffix, suffix, suffix))
    print("{")
    print("    outputFile << setprecision(10) <<")
    print("    detection.metadata_id << \"\\t\" <<")
    print("    detection.object_id << \"\\t\" <<")
    print("    detection.det_id << \"\\t\" <<")

    for row in dataKeys:
        print("    detection.%s << \"\\t\" <<" % (row))

    print("    detection.htm16ID <<")
    print("    endl;")
    print("")
    print("    return 0;")
    print("}")
    print("")
    print("int DetectionCrossmatcher%s::createNewObject(ObjectRow& object, detections%s& detection, metadata%s& metadata, MomentsRow& moments, TransientObjectTableDB& objectDB, detections%sDB& detectionDB, TransientMomentsDB& momentsDB, CommonUtils& common, bool writeFile, bool ingestMoments, ofstream& outputFile, mysqlpp::Query& query)" % (suffix, suffix, suffix, suffix))
    print("{")
    print("    // Create a new object record")
    print("")
    print("    // set the object RA and Dec")
    print("    object.ra = detection.ra;")
    print("    object.dec = detection.dec;")
    print("    object.htm16ID = detection.htm16ID;")
    print("    // Need to get the detection ID before we insert the object.")
    print("    // But need the object ID before we insert the detection.")
    print("    // Thus we need to generate the ID HERE not in the objectDB code.")
    print("    object.object_id = common.generate_ra_dec_id00(object.ra,object.dec);")
    print("    // Insert the detection.")
    print("    detection.object_id = object.object_id;")
    print("")
    print("    long long detection_id = 0;")
    print("    long long moments_id = 0;")
    print("")
    print("    if (writeFile)")
    print("    {")
    print("        int status = writeDetectionRow(outputFile, metadata, detection);")
    print("    }")
    print("    else")
    print("    {")
    print("        detection_id = detectionDB.insertRow(detection, query);")
    print("        // Need to insert moment info here with detection_id")
    print("        if (ingestMoments && detection_id > 0)")
    print("        {")
    print("            moments.detection_id = detection_id;")
    print("            moments_id = momentsDB.insertRow(moments, query);")
    print("        }")
    print("    }")
    print("")
    print("    if (detection_id == 0 && !writeFile)")
    print("    {")
    print("        // Something went wrong - we want the detection insert to always succeed.")
    print("        cerr << \"Failed to insert detection.\" << endl;")
    print("        //return -1;")
    print("    }")
    print("")
    print("    if (ingestMoments)")
    print("    {")
    print("        if (moments_id == 0 && !writeFile)")
    print("        {")
    print("            // Something went wrong - we want the detection insert to always succeed.")
    print("            cerr << \"Failed to insert moments information.\" << endl;")
    print("            //return -1;")
    print("        }")
    print("    }")
    print("")
    print("    object.detection_id = detection_id;")
    print("")
    print("    long long object_id = objectDB.insertRow(object, query);")
    print("    if (object_id == 0 || object_id != object.object_id)")
    print("    {")
    print("        // Something went wrong with the insert of the object.")
    print("        cerr << \"Failed to insert object.\" << endl;")
    print("        // The object may have failed to insert because it already exists! That's fine.")
    print("        //return -1;")
    print("    }")
    print("")
    print("}")
    print("")
    print("int DetectionCrossmatcher%s::checkStarProximity(detections%s& detection, mysqlpp::Query& query, mysqlpp::StoreQueryResult& resultSet, CommonUtils& common, struct FlagsAndRecurrenceRadius frr)" % (suffix, suffix))
    print("{")
    print("    std::list<string> columns;")
    print("    columns.push_back(\"*\");")
    print("    int object_classification = NULL_INT;")
    print("    bool classified = false;")
    print("    //mysqlpp::StoreQueryResult resultSet;")
    print("    double separation;")
    print("    float mag;")
    print("")
    print("    // Near PS1 catalogued star?")
    print("    if (!classified)")
    print("    {")
    print("        cout << \"PS1 star check \" << detection.ra << \" \" << detection.dec << endl;")
    print("        common.htmCircle(16, detection.ra, detection.dec, frr.vStarRadius, \"tcs_cat_ps1_ubercal_stars\", columns, query);")
    print("        resultSet = query.store();")
    print("        query.reset();")
    print("        if (resultSet.num_rows() > 0)")
    print("        {")
    print("            for (size_t j = 0; j < resultSet.num_rows(); j++)")
    print("            {")
    print("                separation = common.calculate_separation(detection.ra, detection.dec, resultSet[j][\"RA\"],resultSet[j][\"Dec\"]);")
    print("                mag = resultSet[j][\"r\"];")
    print("                if (separation < frr.vStarRadius && mag < 17.0)")
    print("                {")
    print("                    classified = true;")
    print("                    object_classification = 2;")
    print("                    cout << \"PS1 match: object_classification = \" << object_classification << endl;")
    print("                    break;")
    print("                }")
    print("            }")
    print("        }")
    print("    }")
    print("")
    print("")
    print("    // Near Guide Star?")
    print("    if (!classified)")
    print("    {")
    print("        cout << \"Guide star check\" << endl;")
    print("        common.htmCircle(16, detection.ra, detection.dec, frr.vStarRadius, \"tcs_guide_star_cat\", columns, query);")
    print("        resultSet = query.store();")
    print("        query.reset();")
    print("        if (resultSet.num_rows() > 0)")
    print("        {")
    print("            for (size_t j = 0; j < resultSet.num_rows(); j++)")
    print("            {")
    print("                separation = common.calculate_separation(detection.ra, detection.dec, resultSet[j][\"RightAsc\"]*common.RAD_TO_DEG_FACTOR,resultSet[j][\"Declination\"]*common.RAD_TO_DEG_FACTOR);")
    print("                if (separation < frr.vStarRadius)")
    print("                {")
    print("                    classified = true;")
    print("                    object_classification = 2;")
    print("                    break;")
    print("                }")
    print("            }")
    print("        }")
    print("    }")
    print("")
    print("")
    print("    // Near 2MASS Star?")
    print("    if (!classified)")
    print("    {")
    print("        cout << \"2MASS star check\" << endl;")
    print("        common.htmCircle(16, detection.ra, detection.dec, frr.vStarRadius, \"tcs_2mass_psc_cat\", columns, query);")
    print("        resultSet = query.store();")
    print("        query.reset();")
    print("        if (resultSet.num_rows() > 0)")
    print("        {")
    print("            for (size_t j = 0; j < resultSet.num_rows(); j++)")
    print("            {")
    print("                separation = common.calculate_separation(detection.ra, detection.dec, resultSet[j][\"ra\"],resultSet[j][\"decl\"]);")
    print("                if (separation < frr.vStarRadius)")
    print("                {")
    print("                    classified = true;")
    print("                    object_classification = 2;")
    print("                    break;")
    print("                }")
    print("            }")
    print("        }")
    print("    }")
    print("")
    print("")
    print("    // Near SDSS DR9 Star?")
    print("    // Skip this check for the time being.")
    print("")
    print("    return object_classification;")
    print("}")
    print("EOF")

def createMain(suffix, headerKeys, dataKeys):
    """createMain.

    Args:
        suffix:
        headerKeys:
        dataKeys:
    """

    print("cat > IngesterMain%s.cpp <<EOF" % (suffix))
    print("//============================================================================")
    print("// Name        : IngesterMain%s.cpp" % (suffix))
    print("// Author      : Ken Smith")
    print("// Version     :")
    print("// Copyright   :")
    print("// Description : Runs a query against a table using the HTM circleRegion code.")
    print("//               Assumes column name is always htm<level>ID")
    print("//============================================================================")
    print("")
    print("#include <cmath>")
    print("#include <mysql++.h>")
    print("")
    print("#include \"DetectionCrossmatcher%s.h\"" % (suffix))
    print("#include \"CommonUtils.h\"")
    print("#include \"Config.h\"")
    print("")
    print("using namespace std;")
    print("")
    print("int main(int argc, char *argv[]) {")
    print("")
    print("    if (argc != 3)")
    print("    {")
    print("        cerr << \"Usage: \" << argv[0] << \" <config file> <%s filename>\" << endl;" % (suffix))
    print("        return -1;")
    print("    }")
    print("")
    print("    string configFile = argv[1];")
    print("    string filename = argv[2];")
    print("")
    print("    DetectionCrossmatcher%s crossmatcher;" % (suffix))
    print("    CommonUtils utils;")
    print("    Config config(configFile);")
    print("")
    print("    map<string, DatabaseMember> dbmap = config.getDatabases();")
    print("    struct FlagsAndRecurrenceRadius frr = config.getFlagsAndRecurrenceRadius();")
    print("")
    print("")
    print("    string fileextension = utils.extractFilenameExtension(filename);")
    print("    if (fileextension != \".%s\")" % (suffix))
    print("    {")
    print("        cerr << \"Please feed me a headed %s file!\" << endl;" % (suffix))
    print("        return -1;")
    print("    }")
    print("")
    print("    if (crossmatcher.crossmatcher(dbmap, frr, filename))")
    print("    {")
    print("        cerr << \"Something went wrong.\" << endl;")
    print("        return -1;")
    print("    }")
    print("")
    print("    return 0;")
    print("}")
    print("EOF")

def createMakefile(suffix):
    """createMakefile.

    Args:
        suffix:
    """
    print("cat > Makefile_%s <<EOF" % (suffix))
    print("HTM_INCLUDE = \$(CODEBASE)/htm/include")
    print("UTILS_INCLUDE = \$(CODEBASE)/ingesters/ps1_transients/cpp/include")
    print("TPHOT_INCLUDE = \$(CODEBASE)/ingesters/tphot/cpp/include")
    print("HTM_LIB = \$(CODEBASE)/htm")
    print("OBJDIR = ./objs")
    print("INCDIR = .")
    print("RM = rm -f")
    print("CC = g++")
    print("DEBUG = -g")
    print("CFLAGS = -Wall -c -fPIC  \$(DEBUG) -Wall -I\$(MYSQL_INCLUDE) -I\$(MYSQLPP_INCLUDE) -I\$(HTM_INCLUDE) -I\$(INCDIR) -I\$(TPHOT_INCLUDE) -Wno-deprecated \$(CLANG_FLAGS) -I\$(UTILS_INCLUDE) -std=c++0x")
    print("LFLAGS = -Wall \$(DEBUG) -L\$(MYSQL_LIB) -L\$(MYSQLPP_LIB) -L\$(HTM_LIB)")
    print("")
    print("LIBS = -lm -lmysqlclient -lmysqlpp -lhtm -lpcrecpp -lyaml-cpp")
    print("")
    print("OBJS = \$(OBJDIR)/IngesterMain%s.o \$(OBJDIR)/DetectionCrossmatcher%s.o \$(OBJDIR)/detections%sDB.o \$(OBJDIR)/metadata%sDB.o \$(OBJDIR)/ReadDDTFile.o \$(OBJDIR)/CommonUtils.o \$(OBJDIR)/Config.o \$(OBJDIR)/TransientMomentsDB.o \$(OBJDIR)/TransientObjectTableDB.o" % (suffix, suffix, suffix, suffix))
    print("")
    print("IngesterMain%s.o: \$(OBJS)" % (suffix))
    print("\t\$(CC)  -o IngesterMain%s \$(OBJS) \$(LFLAGS) \$(LIBS)" % (suffix))
    print("")
    print("\$(OBJDIR)/IngesterMain%s.o: ./IngesterMain%s.cpp" % (suffix, suffix))
    print("\t\$(CC) \$(CFLAGS) ./IngesterMain%s.cpp -o \$@" % (suffix))
    print("")
    print("\$(OBJDIR)/DetectionCrossmatcher%s.o: ./DetectionCrossmatcher%s.cpp" % (suffix, suffix))
    print("\t\$(CC) \$(CFLAGS) ./DetectionCrossmatcher%s.cpp -o \$@" % (suffix))
    print("")
    print("\$(OBJDIR)/detections%sDB.o: ./detections%sDB.cpp" % (suffix, suffix))
    print("\t\$(CC) \$(CFLAGS) ./detections%sDB.cpp -o \$@" % (suffix))
    print("")
    print("\$(OBJDIR)/metadata%sDB.o: ./metadata%sDB.cpp" % (suffix, suffix))
    print("\t\$(CC) \$(CFLAGS) ./metadata%sDB.cpp -o \$@" % (suffix))
    print("")
    print("\$(OBJDIR)/ReadDDTFile.o: \$(CODEBASE)/ingesters/tphot/cpp/src/ReadDDTFile.cpp")
    print("\t\$(CC) \$(CFLAGS) \$(CODEBASE)/ingesters/tphot/cpp/src/ReadDDTFile.cpp -o \$@")
    print("")
    print("\$(OBJDIR)/CommonUtils.o: \$(CODEBASE)/ingesters/ps1_transients/cpp/src/CommonUtils.cpp")
    print("\t\$(CC) \$(CFLAGS) \$(CODEBASE)/ingesters/ps1_transients/cpp/src/CommonUtils.cpp -o \$@")
    print("")
    print("\$(OBJDIR)/Config.o: \$(CODEBASE)/ingesters/tphot/cpp/src/Config.cpp")
    print("\t\$(CC) \$(CFLAGS) \$(CODEBASE)/ingesters/tphot/cpp/src/Config.cpp -o \$@")
    print("")
    print("\$(OBJDIR)/TransientMomentsDB.o: \$(CODEBASE)/ingesters/tphot/cpp/src/TransientMomentsDB.cpp")
    print("\t\$(CC) \$(CFLAGS) \$(CODEBASE)/ingesters/tphot/cpp/src/TransientMomentsDB.cpp -o \$@")
    print("")
    print("\$(OBJDIR)/TransientObjectTableDB.o: \$(CODEBASE)/ingesters/tphot/cpp/src/TransientObjectTableDB.cpp")
    print("\t\$(CC) \$(CFLAGS) \$(CODEBASE)/ingesters/tphot/cpp/src/TransientObjectTableDB.cpp -o \$@")
    print("")
    print("all: IngesterMain%s" % (suffix))
    print("")
    print("clean:")
    print("\t-\$(RM) \$(OBJS) IngesterMain%s" % (suffix))
    print("")
    print("EOF")

def main(argv = None):
    """main.

    Args:
        argv:
    """
    import optparse

    if argv is None:
        argv = sys.argv

    usage = "Usage: %s <inputFile>" % argv[0]
    if len(argv) < 2:
        sys.exit(usage)

    filename = argv[1]
    suffix = filename.split('.')[-1]
    
    hlines, header = readATLASddetHeader(filename, useOrderedDict = True)
    headerKeys = [x.lower() for x in list(header)]
    colData = readGenericDataFile(filename, skipLines=hlines, useOrderedDict = True)
    dataKeys = [x.lower().replace('-','').replace('/','') for x in list(colData[0])]

    # Create database table
    if headerKeys:
        createDBTable('atlas_metadata', suffix, headerKeys)
        print()
        createCPPClass('metadata', suffix, headerKeys)
        print()
        createCPPDBClassH('metadata', suffix, headerKeys)
        print()
        createCPPDBClassCPP('metadata', suffix, headerKeys)

    createDBTable('atlas_detections', suffix, dataKeys)
    print()
    createCPPClass('detections', suffix, dataKeys)
    print()
    createCPPDBClassH('detections', suffix, dataKeys)
    print()
    createCPPDBClassCPP('detections', suffix, dataKeys)

    #print
    createIngesterH(suffix, headerKeys, dataKeys)
    print()
    createIngesterCPP(suffix, headerKeys, dataKeys)

    print()
    createMain(suffix, headerKeys, dataKeys)

    print()
    createMakefile(suffix)



if __name__ == '__main__':
    main()
